import numpy as np
from pyH2A.Utilities.Energy_Conversion import Energy, kWh, eV
from pyH2A.Utilities.input_modification import insert, process_table

class PEC_Plugin:
	'''Simulating H2 production using photoelectrochemical water splitting.

	Parameters
	----------
	Technical Operating Parameters and Specifications > Design Output per Day > Value : float
		Design output in (kg of H2)/day, ``process_table()`` is used.
	PEC Cells > Cell Cost ($/m2) > Value : float
		Cost of PEC cells in $/m2.
	PEC Cells > Lifetime (year) > Value : float
		Lifetime of PEC cells in years before replacement is required.
	PEC Cells > Length (m) > Value : float
		Length of single PEC cell in m.
	PEC Cells > Width (m) > Value : float
		Width of single PEC cell in m.
	Land Area Requirement > Cell Angle (degree) > Value : float
		Angle of PEC cells from the ground, in degrees.
	Land Area Requirement > South Spacing (m) > Value : float
		South spacing of PEC cells in m.
	Land Ares Requirement > East/West Spacing (m) > Value : float
		East/West Spacing of PEC cells in m.
	Solar-to-Hydrogen Efficiency > STH (%) > Value : float
		Solar-to-hydrogen efficiency in percentage or as a value between 0 and 1.
	Solar Input > Mean solar input (kWh/m2/day) > Value : float
		Mean solar input in kWh/m2/day, ``process_table()`` is used.
	Solar Concentrator > Concentration Factor > Value : float, optional
		Concentration factor created by solar concentration module, which is used in combination
		with PEC cells. If "Solar Concentrator" is in dcf.inp, ``process_table()`` is used.

	Returns
	-------
	Non-Depreciable Capital Costs > Land required (acres) > Value : float
		Total land area required in acres.
	Non-Depreciable Capital Costs > Solar Collection Area (m2) > Value : float
		Solar collection area in m2.
	Planned Replacement > Planned Replacement PEC Cells > Cost ($) : float
		Total cost of replacing all PEC cells once.
	Planned Replacement > Planned Replacement PEC Cells > Frequency (years) : float
		Replacement frequency of PEC cells in years, identical to PEC cell lifetime.
	Direct Capital Costs - PEC Cells > PEC Cell Cost ($) > Value : float
		Total cost of all PEC cells.
	PEC Cells > Number > Value : float
		Number of individual PEC cells required for design H2 output capacity.
	'''

	def __init__(self, dcf, print_info):
		if 'Solar Concentrator' in dcf.inp:
			process_table(dcf.inp, 'Solar Concentrator', 'Value')

		process_table(dcf.inp, 'Solar Input', 'Value')
		process_table(dcf.inp, 'Solar-to-Hydrogen Efficiency', 'Value')
		process_table(dcf.inp, 'PEC Cells', 'Value')
		process_table(dcf.inp, 'Land Area Requirement', 'Value')
		process_table(dcf.inp, 'Technical Operating Parameters and Specifications', 'Value')

		self.hydrogen_production(dcf)
		self.PEC_cost(dcf)
		self.land_area(dcf)

		insert(dcf, 'Non-Depreciable Capital Costs', 'Land required (acres)', 'Value', 
		       self.total_land_area_acres, __name__, print_info = print_info)
		insert(dcf, 'Non-Depreciable Capital Costs', 'Solar Collection Area (m2)', 'Value', 
			   self.total_solar_collection_area, __name__, print_info = print_info)
		
		insert(dcf, 'Planned Replacement', 'Planned Replacement PEC Cells', 'Cost ($)', 
			   self.cell_cost, __name__, print_info = print_info)
		insert(dcf, 'Planned Replacement', 'Planned Replacement PEC Cells', 'Frequency (years)', 
			   dcf.inp['PEC Cells']['Lifetime (years)']['Value'], __name__, print_info = print_info)

		insert(dcf, 'Direct Capital Costs - PEC Cells', 'PEC Cell Cost ($)', 'Value', 
			   self.cell_cost, __name__, print_info = print_info)

		insert(dcf, 'PEC Cells', 'Number', 'Value', self.cell_number, __name__, print_info = print_info)

	def hydrogen_production(self, dcf):
		'''Calculation of (kg of H2)/day produced by single PEC cell.
		'''

		pec = dcf.inp['PEC Cells']

		self.cell_area = pec['Length (m)']['Value'] * pec['Width (m)']['Value']
		cell_insolation = Energy(self.cell_area * dcf.inp['Solar Input']['Mean solar input (kWh/m2/day)']['Value'], kWh)

		mol_H2_per_cell = (cell_insolation.J * dcf.inp['Solar-to-Hydrogen Efficiency']['STH (%)']['Value']) / Energy(2*1.229, eV).Jmol
		self.kg_H2_per_cell = (2 * mol_H2_per_cell) / 1000.

	def PEC_cost(self, dcf):
		'''Calculation of cost per cell, number of required cells and total cell cost.
		'''

		cost_per_cell = self.cell_area * dcf.inp['PEC Cells']['Cell Cost ($/m2)']['Value']
		self.cell_number = np.ceil(dcf.inp['Technical Operating Parameters and Specifications']['Design Output per Day']['Value'] / self.kg_H2_per_cell)
		self.cell_cost = self.cell_number * cost_per_cell

	def land_area(self, dcf):
		'''Calculation of total required land area and solar collection area.
		'''

		land = dcf.inp['Land Area Requirement']
		pec = dcf.inp['PEC Cells']

		self.total_solar_collection_area = self.cell_area * self.cell_number

		cell_plan_view = pec['Length (m)']['Value'] * np.cos(np.radians(land['Cell Angle (degree)']['Value']))
		total_length = cell_plan_view + land['South Spacing (m)']['Value']	
		total_width = pec['Width (m)']['Value'] + land['East/West Spacing (m)']['Value']

		self.total_land_area = total_width * total_length * self.cell_number
		self.total_land_area_acres = self.total_land_area * 0.000247105
		


import numpy as np
from functools import lru_cache

from pyH2A.Utilities.input_modification import insert, process_table, read_textfile, file_import

class Hourly_Irradiation_Plugin:
	'''Calculation of hourly and mean daily irradiation data with different module configurations.
	
	Parameters
	----------
	Hourly Irradiation > File > Value : str
		Path to a `.csv` file containing hourly irradiance data as provided by
		https://re.jrc.ec.europa.eu/pvg_tools/en/#TMY, ``process_table()`` is used.
	Irradiance Area Parameters > Module Tilt (degrees) > Value : float
		Tilt of irradiated module in degrees.
	Irradiance Area Parameters > Array Azimuth (degrees) > Value : float
		Azimuth angle of irradiated module in degrees.
	Irradiance Area Parameters > Nominal Operating Temperature (Celcius) : foat
		Nominal operating temperature of irradiated module in degrees Celcius.
	Irradiance Area Parameters > Mismatch Derating > Value : float
		Derating value due to mismatch (percentage or value between 0 and 1).
	Irradiance Area Parameters > Dirt Derating > Value : float
		Derating value due to dirt buildup (percentage or value between 0 and 1).
	Irradiance Area Parameters > Temperature Coefficienct (per Celcius) > Value : float
		Performance decrease of irradiated module per degree Celcius increase.

	Returns
	-------
	Hourly Irradiation > No Tracking (kW) > Value : ndarray
		Hourly irradiation with no tracking per m2 in kW.
	Hourly Irradiation > Horizontal Single Axis Tracking (kW) > Value : ndarray
		Hourly irradiation with single axis tracking per m2 in kW.
	Hourly Irradiation > Two Axis Tracking (kW) > Value : ndarray
		Hourly irradiation with two axis tracking per m2 in kW.
	Hourly Irradiation > Mean solar input (kWh/m2/day) > Value : float
		Mean solar input with no tracking in kWh/m2/day.
	Hourly Irradiation > Mean solar input, single axis tracking (kWh/m2/day) > Value : float
		Mean solar input with single axis tracking in kWh/m2/day.
	Hourly Irradiation > Mean solar input, two axis tracking (kWh/m2/day) > Value : float
		Mean solar input with two axis tracking in kWh/m2/day.
	'''

	def __init__(self, dcf, print_info):

		process_table(dcf.inp, 'Hourly Irradiation', 'Value')

		data, location = import_hourly_data(dcf.inp['Hourly Irradiation']['File']['Value'])

		insert(dcf, 'Hourly Irradiation', 'Latitude', 'Value', 
			   location['Latitude (decimal degrees)'], __name__, print_info = print_info)
		insert(dcf, 'Hourly Irradiation', 'Longitude', 'Value', 
			   location['Longitude (decimal degrees)'], __name__, print_info = print_info)

		process_table(dcf.inp, 'Irradiance Area Parameters', 'Value')

		pv = dcf.inp['Irradiance Area Parameters']

		self.power_kW, self.power_sat_kW, self.power_dat_kW = calculate_PV_power_ratio(dcf.inp['Hourly Irradiation']['File']['Value'],
											pv['Module Tilt (degrees)']['Value'], pv['Array Azimuth (degrees)']['Value'],
											pv['Nominal Operating Temperature (Celcius)']['Value'], 
											pv['Temperature Coefficienct (per Celcius)']['Value'],
											pv['Mismatch Derating']['Value'], pv['Dirt Derating']['Value'])

		insert(dcf, 'Hourly Irradiation', 'No Tracking (kW)', 'Value', 
			   self.power_kW, __name__, print_info = print_info)
		insert(dcf, 'Hourly Irradiation', 'Horizontal Single Axis Tracking (kW)', 'Value', 
			   self.power_sat_kW, __name__, print_info = print_info)
		insert(dcf, 'Hourly Irradiation', 'Two Axis Tracking (kW)', 'Value', 
			   self.power_dat_kW, __name__, print_info = print_info)

		insert(dcf, 'Hourly Irradiation', 'Mean solar input no tracking (kWh/m2/day)', 'Value', 
			   np.sum(self.power_kW)/365., __name__, print_info = print_info)
		insert(dcf, 'Hourly Irradiation', 'Mean solar input single axis tracking (kWh/m2/day)', 'Value', 
			   np.sum(self.power_sat_kW)/365., __name__, print_info = print_info)
		insert(dcf, 'Hourly Irradiation', 'Mean solar input two axis tracking (kWh/m2/day)', 'Value', 
			   np.sum(self.power_dat_kW)/365., __name__, print_info = print_info)

def converter_function(string):
	'''Converter function for datetime of hourly irradiation data.'''

	decoded = string.decode('utf-8')
	split = decoded.split(':')

	return float(split[1][:2]) #- 0.5

def import_Chang_data(file_name):
	'''Import of Chang 2020 data, for debugging.'''

	file_name = 'pyH2A.Lookup_Tables.Hourly_Irradiation_Data~Hourly_Irradiation_Data_Townsville_Chang_2020.csv'

	data = read_textfile(file_name, delimiter = '	')

	data_dict = {'Time': data[:,0] - 10., 'Temperature': data[:,1],
				  'Direct Normal Irradiance': data[:,2], 'Diffuse Horizontal Irradiance': data[:,3]}

	location = {'Latitude (decimal degrees)': -19.25, 'Longitude (decimal degrees)': 146.77}

	return data_dict, location
	
@lru_cache(maxsize = None)
def import_hourly_data(file_name):
	'''Imports hourly irradiation data and location coordinates from the `.csv` format provided 
	by: https://re.jrc.ec.europa.eu/pvg_tools/en/#TMY.
	``@lru_cache`` is used for fast repeated reads
	'''

	data = np.genfromtxt(file_import(file_name, mode = 'r'), 
						  delimiter = ',', skip_header = 17, 
						  skip_footer = 9, converters = {0: converter_function})

	strings = ['Latitude (decimal degrees)', 'Longitude (decimal degrees)']
	location = {}

	file_read = file_import(file_name, mode = 'r')
	for row_counter, line in enumerate(file_read):

		split = line.split(':')

		if split[0] in strings:
			location[split[0]] = float(split[1].strip(' '))
		else:
			break
	file_read.close()

	data_dict = {'Time': data[:,0], 'Temperature': data[:,1], 'Global Horizontal Irradiance':  data[:,3],
				 'Direct Normal Irradiance': data[:,4], 'Diffuse Horizontal Irradiance': data[:,5]}

	return data_dict, location

@lru_cache(maxsize = None)
def calculate_PV_power_ratio(file_name, module_tilt, array_azimuth, nominal_operating_temperature,
							 temperature_coefficient, mismatch_derating, dirt_derating):
	'''Calculation based on Chang 2020, https://doi.org/10.1016/j.xcrp.2020.100209
	SAT: horzontal single axis tracking
	DAT: dual axis tracking, no diffuse radiation
	'''

	data, location = import_hourly_data(file_name)
	#data, location = import_Chang_data(file_name)

	latitude = location['Latitude (decimal degrees)']
	longitude = location['Longitude (decimal degrees)']

	day_number = np.arange(1, len(data['Time']) + 1) / 24
	#day_number = np.arange(0, len(data['Time'])) / 24

	declination_angle = 23.45 * np.sin((day_number - 81) * 2 * np.pi / 365.)
	hour_angle = (data['Time'] - 12) * 15 + longitude

	altitude_angle = 360 / (2 * np.pi) * np.arcsin(np.sin(2 * np.pi / 360 * declination_angle) * 
					 np.sin(2 * np.pi / 360 * latitude) + np.cos(2 * np.pi / 360 * declination_angle) * 
					 np.cos(2 * np.pi / 360 * latitude) * np.cos(2 * np.pi / 360 * hour_angle))

	azimuth_angle = 360 / (2 * np.pi) * np.arccos((np.sin(2 * np.pi / 360 * declination_angle) * 
					np.cos(2 * np.pi / 360 * latitude) - np.cos( 2 * np.pi / 360 * declination_angle) * 
					np.sin(2 * np.pi / 360 * latitude) * np.cos(2 * np.pi / 360 * hour_angle)) / 
					np.cos(2 * np.pi / 360 * altitude_angle)) * np.sign(hour_angle)

	dni_fraction = np.cos(2 * np.pi / 360 * altitude_angle) * np.sin(2 * np.pi / 360 * 
				   module_tilt) * np.cos(2 * np.pi / 360 * (array_azimuth - 
				   azimuth_angle)) + np.sin(2 * np.pi / 360 * altitude_angle) * np.cos(2 * np.pi / 
				   360 * module_tilt)
	dni_fraction = dni_fraction.clip(min = 0)

	direct_plane_radiation = data['Direct Normal Irradiance'] * dni_fraction
	diffuse_plane_radiation = data['Diffuse Horizontal Irradiance'] * (180 - module_tilt) / 180
	total_plane_radiation = direct_plane_radiation + diffuse_plane_radiation

	cell_temperature = data['Temperature'] + (nominal_operating_temperature - 
					   20) * total_plane_radiation/800  #where does this formula come from?

	temperature_derating = 1 + temperature_coefficient * (cell_temperature - 25)  # why the 25 correction?

	power_kW = (temperature_derating * mismatch_derating * 
					 dirt_derating * total_plane_radiation/1000)  # Converting W to kW

	sat_azimuth = np.sign(azimuth_angle) * 90

	sat_tilt = 360 / (2 * np.pi) * np.arctan(1 / np.tan(2 * np.pi / 360 * altitude_angle) * 
			   np.cos( 2 * np.pi / 360 * (sat_azimuth - azimuth_angle)))

	sat_fraction = (np.cos(2 * np.pi / 360 * altitude_angle) * np.sin(2 * np.pi / 360 * sat_tilt) * 
					np.cos(2 * np.pi / 360 * (sat_azimuth - azimuth_angle)) + np.sin(2 * np.pi / 360 * altitude_angle) * 
					np.cos(2 * np.pi / 360 * sat_tilt))
	sat_fraction = sat_fraction.clip(min = 0)

	sat_direct_POA = sat_fraction * data['Direct Normal Irradiance']
	sat_diffuse_POA = data['Diffuse Horizontal Irradiance'] * (180 - sat_tilt) / 180
	sat_total_POA = sat_direct_POA + sat_diffuse_POA

	power_sat_kW = (temperature_derating * mismatch_derating * 
					 dirt_derating * sat_total_POA / 1000)  # Convert W to kW

	power_dat_kW = (data['Direct Normal Irradiance'] * temperature_derating * 
					mismatch_derating * dirt_derating / 1000)

	return power_kW, power_sat_kW, power_dat_kW




import click
from pyH2A.run_pyH2A import command_line_pyH2A

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
	pass

@cli.command()
@click.option('-i', '--input_file', type=str, help='Path to input file.')
@click.option('-o', '--output_dir', type=str, help='Path to output directory.')
def run(input_file, output_dir):
	'''Run pyH2A analysis.
	'''
	output = command_line_pyH2A(input_file, output_dir)





import numpy as np
from pyH2A.Utilities.input_modification import insert, process_table

class Multiple_Modules_Plugin:
	''' Simulating mutliple plant modules which are operated together, assuming that only labor cost is reduced. 
	Calculation of required labor to operate all modules, scaling down labor requirement to one module for subsequent calculations.

	Parameters
	----------
	Technical Operating Parameters and Specification > Plant modules > Value : float or int
		Number of plant modules considered in this calculation, ``process_table()`` is used.
	Non-Depreciable Capital Costs > Solar Collection Area (m2) > Value : float
		Solar collection area for one plant module in m2, ``process_table()`` is used.
	Fixed Operating Costs > area > Value : float
		Solar collection area in m2 that can be covered by one staffer.
	Fixed Operating Costs > shifts > Value : float or int
		Number of 8-hour shifts (typically 3 for 24h operation).
	Fixed Operating Costs > supervisor > Value : float or int
		Number of shift supervisors.

	Returns
	-------
	Fixed Operating Costs > staff > Value : float
		Number of 8-hour equivalent staff required for operating one plant module.
	''' 

	def __init__(self, dcf, print_info):
		process_table(dcf.inp, 'Technical Operating Parameters and Specifications', 'Value')
		process_table(dcf.inp, 'Non-Depreciable Capital Costs', 'Value')
		process_table(dcf.inp, 'Fixed Operating Costs', 'Value')

		self.required_staff(dcf)

		insert(dcf, 'Fixed Operating Costs', 'staff', 'Value', self.staff_per_module, __name__, print_info = print_info)

	def required_staff(self, dcf):
		'''Calculation of total required staff for all plant modules, then scaling down to staff
		requirements for one module.'''

		area = dcf.inp['Technical Operating Parameters and Specifications']['Plant Modules']['Value'] * dcf.inp['Non-Depreciable Capital Costs']['Solar Collection Area (m2)']['Value']

		staff = np.ceil(area / dcf.inp['Fixed Operating Costs']['area']['Value']) + dcf.inp['Fixed Operating Costs']['supervisor']['Value']
		staff = staff * dcf.inp['Fixed Operating Costs']['shifts']['Value']

		self.staff_per_module = staff / dcf.inp['Technical Operating Parameters and Specifications']['Plant Modules']['Value']


import math
from pathlib import PurePath
from pathlib import Path
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import matplotlib.image as mpimg
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
import numpy as np
from pyH2A.Utilities.input_modification import file_import

def make_bold(string):
	'''Convert provided string to a string which is formatted to be bold.
	"%" signs cannot be rendered bold with this approach and are hence returned normal'''

	if isinstance(string, str):
		string = string.split(' ')
	elif isinstance(string, float):
		string = [str(string)]

	output = ''

	for word in string:
		if '%' in word:
			output += r'$\bf{' + word[:-1] + '}$' + word[-1] + ' '
		else:
			output += r'$\bf{' + word + '}$' + ' '

	return output

def set_font(font_family, font, font_size):
	'''Set font for plot.

	Parameters
	----------
	font_family : str
		Font family, either 'serif' or 'sans-serif'.
	font : str
		Name of font.
	font_size : float
		Font size.
	'''

	from matplotlib import rcParams

	rcParams['font.family'] = font_family
	rcParams['font.sans-serif'] = [font]
	rcParams['font.size'] = font_size

class Figure_Lean:
	'''Wrapper class for figures.

	Parameters
	----------
	name : str
		Name of figure, used for saving.
	directory : str
		Path to directory where figure is to be saved.
	provided_figure_and_axis : tuple or None, optional
		Tuple of matplotlib.fig and matplotlib.ax objects. If ``None`` new `fig`
		and `ax` objects are generated.
	show : bool, optional
		If True, figure is shown.
	save : bool, optional
		If True, figure is saved.
	pdf : bool, optional
		If True, figure is saved as a PDF file. If False, it is saved as a PNG
		file.
	dpi : int, optional
		Dots per inch resolution of figure.
	transparent : bool, optional
		Flag to control if background of figure is transparent or not.
	nrows : int, optional
		Number of rows for subplots.
	ncols : int, optional
		Number of columns for suplots.
	fig_width : float, optional
		Width of figure in inches.
	fig_height : float, optional
		Height of figure in inches.
	left : float, optional
		Left edge of plot in axis fraction coordinates.
	right : float, optional
		Right edge of plot in axis fraction coordinates.
	top : float, optional
		Top edge of plot in axis fraction coordinates.
	bottom : float, optional
		Bottom edge of plot in axis fraction coordinates.
	wspace : float, optional
		Vertical white space between subplots.
	hspace : float, optional
		Horizontal white space between subplots.
	font_family : str, optional
		Font family, either 'serif' or 'sans-serif'.
	font : str, optional
		Name of font.
	font_size : float, optional
		Font size.
	sharex : bool, optional
		Flag to control if x axis is shared between subplots.

	Notes
	-----
	Provided figure is shown and/or saved in provided directory with given name by
	running `Figure_Lean.execute()`.
	'''
	def __init__(self, name, directory, provided_figure_and_axis = None, 
				 show = False, save = False, 
				 pdf = False, dpi = 150, transparent = False,
				 nrows = 1, ncols = 1, fig_width = 6.4, fig_height = 4.8, 
				 left = 0.125, right = 0.9, top = 0.88, bottom = 0.11, 
				 wspace = 0.2, hspace = 0.2, font_family = 'sans-serif',
				 font = 'Arial', font_size = 12, sharex = False):

		set_font(font_family, font, font_size)

		if provided_figure_and_axis is None:
			fig, ax = plt.subplots(nrows = nrows, ncols = ncols, 
								   figsize = (fig_width, fig_height), sharex = sharex)
		else:
			fig, ax = provided_figure_and_axis

		fig.subplots_adjust(left = left, right = right, top = top, bottom = bottom, 
							wspace = wspace, hspace = hspace)


		self.fig = fig
		self.ax = ax

		self.name = name
		self.directory = directory

		self.show = show
		self.save = save
		self.pdf = pdf
		self.dpi = dpi
		self.transparent = transparent

	def execute(self):
		'''Running `self.execute()` executes desired 
		`show` and `save` options.
		'''
		if self.show:
			plt.show()
		else:
			plt.close()

		if self.save:
			self.save_figure(self.pdf, self.dpi, self.transparent)

	def save_figure(self, pdf, dpi, transparent):
		'''Saving figure in target dictionary with specified 
		parameters.
		'''

		if pdf is True:
			suffix = '.pdf'
		else:
			suffix = '.png'

		path_to_file = PurePath(self.directory, self.name + suffix)

		self.fig.savefig(path_to_file, 
						 transparent = transparent, 
						 dpi = dpi)

		plt.close()

def millify(n, dollar_sign = True):
	'''Converts n to a string with shorthand notation for thousand-steps'''

	millnames = ['','K','M','B','T']
	n = float(n)
	millidx = max(0,min(len(millnames)-1, int(math.floor(0 if n == 0 else math.log10(abs(n))/3))))

	if dollar_sign:
		return '${:.2f}{}'.format(n / 10**(3 * millidx), millnames[millidx])
	else:
		return '{:.2f}{}'.format(n / 10**(3 * millidx), millnames[millidx])

class MathTextSciFormatter(mticker.Formatter):
	'''Formatter for scientific notation in MathText.

	Methods
	-------
	__call__:
		Call method.
	fix_minus:
		Fixing minus.
	format_data:
		Format data method.
	format_data_short:
		Format data shortened method.
	format_ticks:
		Format ticks methods.
	'''

	def __init__(self, fmt="%1.1e"):
	    self.fmt = fmt
	def __call__(self, x, pos=None):
		s = self.fmt % x
		decimal_point = '.'
		positive_sign = '+'
		tup = s.split('e')
		significand = tup[0].rstrip(decimal_point)
		sign = tup[1][0].replace(positive_sign, '')
		exponent = tup[1][1:].lstrip('0')
		if exponent:
			exponent = '10^{%s%s}' % (sign, exponent)
		if significand and exponent:
			s =  r'%s{\times}%s' % (significand, exponent)
		else:
			s =  r'%s%s' % (significand, exponent)
		return "${}$".format(s)

def format_scientific(value):
	'''Converts value to string with scientfic (10**x) notation'''

	formatter = MathTextSciFormatter()
	return formatter(value)

def dynamic_value_formatting(value, cutoff = 6):
	'''Dynamic formatiing of value to string.

	Parameters
	----------
	cutoff : int, optional
		Cutoff value for string length. Below cutoff value a
		string is shown without special formatting.

	Notes
	-----
	If value is an int (or a float that can be represented as an int) and its 
	length as a string is less than the `cutoff value`, it will be printed as such.
	If its length as a string is more than the cutoff value, it will be either printed 
	using the `millify` function (if the value is larger than 1), or using the 
	`format_scientific` function (if the value is smaller than 1).
	'''

	if isinstance(value, int) or isinstance(value, np.int64):
		pass
	elif isinstance(value, float):
		if value.is_integer():
			value = int(value)
	else:
		raise ValueError('{0} is non-numeric. type = {1}'.format(value, type(value)))

	if len(str(value)) < cutoff:
		return str(value)
	else:

		if value > 1:
			return millify(value, dollar_sign = False)
		else:
			return format_scientific(value)

def bottom_offset(self, bboxes, bboxes2):
	'''Bottom offset for cost contribution plot labels.
	'''

	pad = plt.rcParams["xtick.major.size"] + plt.rcParams["xtick.major.pad"]
	
	bottom = self.axes.bbox.ymin
	self.offsetText.set(va="top", ha="left") 
	oy = bottom - pad * self.figure.dpi / 72.0
	self.offsetText.set_position((1.02, oy))

def insert_image(path, x, y, zoom, ax):
	'''Insert image into plot.

	Parameters
	----------
	path : str
		Path to image to be inserted.
	x : float
		x axis coordinate of image in axis fraction coordinates.
	y : float
		y axis coordinate of image in axis fraction coordinates.
	ax : matplotlib.ax
		matplotlib.ax object into which image is inserted.
	'''

	img = mpimg.imread(file_import(path))
	imagebox = OffsetImage(img, zoom = zoom)
	ab = AnnotationBbox(imagebox, (x, y), frameon = False, xycoords = ax.transAxes)
	ax.add_artist(ab)




from pyH2A.Utilities.input_modification import insert, sum_all_tables, process_table

class Fixed_Operating_Cost_Plugin:
	'''Calculation of yearly fixed operating costs.

	Parameters
	----------
	Fixed Operating Costs > staff > Value : float or int
		Number of staff, ``process_table()`` is used.
	Fixed Operating Costs > hourly labor cost > value : float
		Hourly labor cost of staff, ``process_table()`` is used.
	[...] Other Fixed Operating Cost [...] >> Value : float
		Yearly other fixed operating costs, ``sum_all_tables()`` is used.

	Returns
	-------
	[...] Other Fixed Operating Cost [...] > Summed Total > Value : float
		Summed total for each individual table in "Other Fixed Operating Cost" group.
	Fixed Operating Costs > Labor Cost - Uninflated > Value : float
		Yearly total labor cost.
	Fixed Operating Costs > Labor Cost > Value : float
		Yearly total labor cost multiplied by labor inflator.
	Fixed Operating Costs > Total > Value : float
		Sum of total yearly labor costs and yearly other fixed operating costs.
	'''


	def __init__(self, dcf, print_info):
		self.labor_cost(dcf)
		insert(dcf, 'Fixed Operating Costs', 'Labor Cost - Uninflated', 'Value', self.labor_uninflated, __name__, print_info = print_info)
		insert(dcf, 'Fixed Operating Costs', 'Labor Cost', 'Value', self.labor, __name__, print_info = print_info)

		self.other_cost(dcf, print_info)
		insert(dcf, 'Fixed Operating Costs', 'Total', 'Value', self.labor + self.other, __name__, print_info = print_info)

	def labor_cost(self, dcf):
		'''Calculation of yearly labor costs by multiplying number of staff times hourly labor cost.'''

		process_table(dcf.inp, 'Fixed Operating Costs', 'Value')

		self.labor_uninflated = dcf.inp['Fixed Operating Costs']['staff']['Value'] * dcf.inp['Fixed Operating Costs']['hourly labor cost']['Value'] * 2080.
		self.labor = self.labor_uninflated * dcf.labor_inflator 
	
	def other_cost(self, dcf, print_info):
		'''Calculation of yearly other fixed operating costs by applying ``sum_all_tables()`` 
		to "Other Fixed Operating Cost" group.'''

		self.other = sum_all_tables(dcf.inp, 'Other Fixed Operating Cost', 'Value', insert_total = True, class_object = dcf, print_info = print_info) * dcf.combined_inflator




import numpy as np
from pyH2A.Utilities.Energy_Conversion import Energy, kWh, eV
from pyH2A.Utilities.input_modification import insert, process_table

class Photocatalytic_Plugin:
	'''Simulating H2 production using photocatalytic water splitting in plastic baggie reactors.

	Parameters
	----------
	Technical Operating Parameters and Specifications > Design Output per Day > Value : float
		Design output in (kg of H2)/day, ``process_table()`` is used.
	Reactor Baggies > Cost Material Top ($/m2) > Value : float
		Cost of baggie top material in $/m2.
	Reactor Baggies > Cost Material Bottom ($/m2) > Value : float
		Cost of baggie bottom material in $/m2.
	Reactor Baggies > Number of ports > Value : int
		Number of ports per baggie.
	Reactor Baggies > Other Costs ($) > Value : float
		Other costs per baggie.
	Reactor Baggies > Markup factor > Value : float
		Markup factor for baggies, typically > 1.
	Reactor Baggies > Length (m) > Value : float
		Length of single baggie in m.
	Reactor Baggies > Width (m) > Value : float
		Width of single baggie in m.
	Reactor Baggies > Height (m) > Value : float
		Height of reactor baggie in m. In this simulation this value determines the height
		of the water level and hence is an important parameter ultimately determining the
		level of light absorption and total catalyst amount.
	Reactor Baggies > Additional land area (%) > Value : float
		Additional land area required, percentage or value > 0. 
		Calculated as: (1 + addtional land area) * baggie area.
	Reactor Baggies > Lifetime (years) > Value : float
		Lifetime of reactor baggies in years before replacement is required.
	Catalyst > Cost per kg ($) > Value : float
		Cost per kg of catalyst.
	Catalyst > Concentration (g/L) > Value : float
		Concentration of catalyst in g/L.
	Catalyst > Lifetime (years) > Value : float
		Lifetime of catalysts in year before replacement is required.
	Catalyst > Molar Weight (g/mol) > Value : float, optional
		If the molar weight of the catalyst (in g/mol) is specified, homogeneous catalyst
		properties (TON, TOF etc. are calculated).
	Catalyst > Molar Attenuation Coefficient (M^-1 cm^-1) > Value : float, optional
		If the molar attenuation coefficient (in M^-1 cm^-1) is specified (along with the molar weight),
		absorbance and the fraction of absorbed light are also calculated.
	Solar-to-Hydrogen Efficiency > STH (%) > Value : float
		Solar-to-hydrogen efficiency in percentage or as a value between 0 and 1.
	Solar Input > Mean solar input (kWh/m2/day) > Value : float
		Mean solar input in kWh/m2/day, ``process_table()`` is used.

	Returns
	-------
	Non-Depreciable Capital Costs > Land required (acres) > Value : float
		Total land area required in acres.
	Non-Depreciable Capital Costs > Solar Collection Area (m2) > Value : float
		Solar colelction area in m2.
	Planned Replacement > Planned Replacement Catalyst > Cost ($) : float
		Total cost of completely replacing the catalyst once.
	Planned Replacement > Planned Replacement Catalyst > Frequency (years) : float
		Replacement frequency of catalyst in years, identical to catalyst lifetime.
	Planned Replacement > Planned Replacement Baggie > Cost ($) : float
		Total cost of replacing all  baggies.
	Planned Replacement > Planned Replacement Baggie > Frequency (years) : float
		Replacement frequency of baggies in year, identical to baggie lifetime.
	Direct Capital Costs - Reactor Baggies > Baggie Cost ($) > Value : float
		Total baggie cost.
	Direct Capital Costs - Photocatalyst > Catalyst Cost ($) > Value : float
		Total catalyst cost.
	Reactor Baggies > Number > Value : int
		Number of individual baggies required for design H2 production capacity.
	Catalyst > Properties > Value : dict
		Dictionary containing detailed catalyst properties calculated from provided parameters.
	['Photocatalytic_Plugin'].catalyst_properties : dict
		Attribute containing catalyst properties dictionary.
	'''

	def __init__(self, dcf, print_info):
		process_table(dcf.inp, 'Reactor Baggies', 'Value')
		process_table(dcf.inp, 'Solar Input', 'Value')
		process_table(dcf.inp, 'Solar-to-Hydrogen Efficiency', 'Value')

		self.hydrogen_production(dcf)

		process_table(dcf.inp, 'Technical Operating Parameters and Specifications', 'Value')

		self.baggie_cost(dcf)

		process_table(dcf.inp, 'Catalyst', 'Value')

		self.catalyst_cost(dcf)
		self.land_area(dcf)

		self.catalyst_activity(dcf)

		insert(dcf, 'Non-Depreciable Capital Costs', 'Land required (acres)', 'Value', 
				self.total_land_area_acres, __name__, print_info = print_info)
		insert(dcf, 'Non-Depreciable Capital Costs', 'Solar Collection Area (m2)', 'Value', 
				self.total_solar_collection_area, __name__, print_info = print_info)
		
		insert(dcf, 'Planned Replacement', 'Planned Replacement Catalyst', 'Cost ($)', 
				self.catalyst_cost, __name__, print_info = print_info)
		insert(dcf, 'Planned Replacement', 'Planned Replacement Catalyst', 'Frequency (years)', 
				dcf.inp['Catalyst']['Lifetime (years)']['Value'], __name__, print_info = print_info)
		insert(dcf, 'Planned Replacement', 'Planned Replacement Baggie', 'Cost ($)', 
				self.baggies_cost, __name__, print_info = print_info)
		insert(dcf, 'Planned Replacement', 'Planned Replacement Baggie', 'Frequency (years)', 
				dcf.inp['Reactor Baggies']['Lifetime (years)']['Value'], __name__, print_info = print_info)

		insert(dcf, 'Direct Capital Costs - Reactor Baggies', 'Baggie Cost ($)', 'Value', 
				self.baggies_cost, __name__, print_info = print_info)
		insert(dcf, 'Direct Capital Costs - Photocatalyst', 'Catalyst Cost ($)', 'Value', 
				self.catalyst_cost, __name__, print_info = print_info)

		insert(dcf, 'Reactor Baggies', 'Number', 'Value', self.baggie_number, __name__, print_info = print_info)
		insert(dcf, 'Catalyst', 'Properties', 'Value', self.catalyst_properties, __name__, print_info = print_info)

	def hydrogen_production(self, dcf):
		'''Calculation of hydrogen produced per day per baggie (in kg).
		'''

		baggie = dcf.inp['Reactor Baggies']

		self.baggie_area = baggie['Length (m)']['Value'] * baggie['Width (m)']['Value']
		baggie_insolation = Energy(self.baggie_area * dcf.inp['Solar Input']['Mean solar input (kWh/m2/day)']['Value'], kWh)

		mol_H2_per_baggie = (baggie_insolation.J * 
					dcf.inp['Solar-to-Hydrogen Efficiency']['STH (%)']['Value']) / Energy(2*1.229, eV).Jmol

		self.kg_H2_per_baggie = (2 * mol_H2_per_baggie)/1000.

	def catalyst_activity(self, dcf):
		'''Calculation of detailed catalyst properties based on provided parameters. If "Molar Weight (g/mol)"
		is specified in "Catalyst" table properties of a homogeneous catalyst are also calculated. Furthermore,
		if "Molar Attenuation Coefficient (M^-1 cm^-1)" is also provided, the light absorption properties 
		are calculated.
		'''

		catalyst_properties = {}

		peak_hourly_irradiation_per_m2 = np.amax(dcf.inp['Solar Input']['Hourly (kWh/m2)']['Value'])
		peak_hourly_irradiation_per_m2 = Energy(peak_hourly_irradiation_per_m2, kWh)

		peak_mol_H2_per_m2_per_h = (peak_hourly_irradiation_per_m2.J * 
								dcf.inp['Solar-to-Hydrogen Efficiency']['STH (%)']['Value']) / Energy(2 * 1.229, eV).Jmol

		mean_daily_mol_H2_per_m2 = (Energy(dcf.inp['Solar Input']['Mean solar input (kWh/m2/day)']['Value'], kWh).J *
									dcf.inp['Solar-to-Hydrogen Efficiency']['STH (%)']['Value']) / Energy(2 * 1.229, eV).Jmol

		kg_catalyst_per_m2 = ((dcf.inp['Reactor Baggies']['Height (m)']['Value'] * 1000) * 
								dcf.inp['Catalyst']['Concentration (g/L)']['Value']/1000.)

		activity_mmol_H2_per_h_per_g_catalyst = 1000 * peak_mol_H2_per_m2_per_h / (kg_catalyst_per_m2 * 1000)

		catalyst_properties['Peak activity / mmol H2/h/g'] = activity_mmol_H2_per_h_per_g_catalyst
		catalyst_properties['Peak H2 production / mol H2/m2/h'] = peak_mol_H2_per_m2_per_h
		catalyst_properties['Catalyst Conc. / kg/m2'] = kg_catalyst_per_m2
		catalyst_properties['Catalyst Conc. / g/L'] = dcf.inp['Catalyst']['Concentration (g/L)']['Value']
	
		if 'Molar Weight (g/mol)' in dcf.inp['Catalyst']:

			catalyst_mol_per_L = ((dcf.inp['Catalyst']['Concentration (g/L)']['Value']) /
								   dcf.inp['Catalyst']['Molar Weight (g/mol)']['Value'])

			liter_per_m2 = dcf.inp['Reactor Baggies']['Height (m)']['Value'] * 1000

			mol_catalyst_per_m2 = liter_per_m2 * catalyst_mol_per_L

			peak_TOF_hourly = peak_mol_H2_per_m2_per_h / mol_catalyst_per_m2
			average_TOF_daily = mean_daily_mol_H2_per_m2 / mol_catalyst_per_m2
			TON = average_TOF_daily * dcf.inp['Catalyst']['Lifetime (years)']['Value'] * 365

			catalyst_properties['Homogeneous'] = {}
			catalyst_properties['Homogeneous']['Catalyst Conc. / mol/L'] = catalyst_mol_per_L
			catalyst_properties['Homogeneous']['Catalyst Conc. / mol/m2'] = mol_catalyst_per_m2
			catalyst_properties['Homogeneous']['Peak TOF / h^-1'] = peak_TOF_hourly
			catalyst_properties['Homogeneous']['Mean daily TOF / d^-1'] = average_TOF_daily
			catalyst_properties['Homogeneous']['TON'] = TON

			if 'Molar Attenuation Coefficient (M^-1 cm^-1)' in dcf.inp['Catalyst']:
				absorbance = (catalyst_mol_per_L * (dcf.inp['Reactor Baggies']['Height (m)']['Value'] * 100) * 
						dcf.inp['Catalyst']['Molar Attenuation Coefficient (M^-1 cm^-1)']['Value'])

				catalyst_properties['Homogeneous']['Absorbance'] = absorbance
				catalyst_properties['Homogeneous']['Absorbed light (%)'] = 100 * (1 - 10**(-absorbance))

			kg_H2_per_day_TOF_calculation = 1000 * self.catalyst_amount / dcf.inp['Catalyst']['Molar Weight (g/mol)']['Value'] * average_TOF_daily * 2. / 1000.
			kg_H2_per_day_baggie_calculation = self.kg_H2_per_baggie * self.baggie_number

			assert abs(kg_H2_per_day_TOF_calculation - kg_H2_per_day_baggie_calculation) < 1e-6, 'Difference between baggie and TOF calculation for daily H2 production: TOF: {0}, Baggie: {0}.'.format(
					kg_H2_per_day_TOF_calculation, kg_H2_per_day_baggie_calculation)

		self.catalyst_properties = catalyst_properties

	def baggie_cost(self, dcf):
		'''Calculation of cost per baggie, number of required baggies and total baggie cost.
		'''

		baggie = dcf.inp['Reactor Baggies']

		material_cost = self.baggie_area * (baggie['Cost Material Top ($/m2)']['Value'] + baggie['Cost Material Bottom ($/m2)']['Value'])
		port_cost = baggie['Number of ports']['Value'] * baggie['Cost of port ($)']['Value']

		cost_per_baggie = baggie['Markup factor']['Value'] * (material_cost + port_cost + baggie['Other Costs ($)']['Value'])

		self.baggie_number = np.ceil(dcf.inp['Technical Operating Parameters and Specifications']['Design Output per Day']['Value'] / self.kg_H2_per_baggie)
		self.baggies_cost = self.baggie_number * cost_per_baggie

	def catalyst_cost(self, dcf):
		'''Calculation of individual baggie volume, catalyst amount per baggie, total catalyst amount 
		and total catalyst cost.
		'''

		baggie = dcf.inp['Reactor Baggies']

		baggie_volume_m3 = baggie['Length (m)']['Value'] * baggie['Width (m)']['Value'] * baggie['Height (m)']['Value']
		baggie_volume_liters = baggie_volume_m3 * 1000

		self.catalyst_amount_per_baggie = baggie_volume_liters * dcf.inp['Catalyst']['Concentration (g/L)']['Value']/1000.
		self.catalyst_amount = self.catalyst_amount_per_baggie * self.baggie_number

		self.catalyst_cost = self.catalyst_amount * dcf.inp['Catalyst']['Cost per kg ($)']['Value']

	def land_area(self, dcf):
		'''Calculation of total required land area and solar collection area.
		'''

		baggie_land_area = self.baggie_number * self.baggie_area
		total_land_area = baggie_land_area * (1. + dcf.inp['Reactor Baggies']['Additional land area (%)']['Value'])

		self.total_land_area_acres = total_land_area * 0.000247105
		self.total_solar_collection_area = baggie_land_area



import sys
import ast
from pyH2A.Discounted_Cash_Flow import Discounted_Cash_Flow
from pyH2A.Utilities.input_modification import convert_input_to_dictionary, execute_plugin, convert_dict_to_kwargs_dict #, parse_file_name

from timeit import default_timer as timer

class pyH2A:
	'''pyH2A class that performs discounted cash flow analysis and executes analysis modules.

	Parameters
	----------
	input_file : str
		Path to input file.
	output_directory : str
		Path to output file.
	print_info : bool, optional
		Flag to control if detailed information during run of pyH2A is printed.

	Returns
	-------
	pyH2A : object
		pyH2A class object.

	Attributes
	----------
	inp : dict
		Dictionary containing input information and all information from discounted
		cash flow analysis.
	base_case : Discounted_Cash_Flow object
		Discounted_Cash_Flow object for base case defined in input file with corresponding
		attributes.
	meta_modules : dict
		Dictionary containing class instances of executes analysis modules.
	'''

	def __init__(self, input_file, output_directory, print_info = False):
		self.input_file = input_file
		self.output_directory = output_directory
		#self.file_name = parse_file_name(self.input_file)
		self.inp = convert_input_to_dictionary(self.input_file)
		self.base_case = Discounted_Cash_Flow(self.input_file, print_info = print_info)

		self.meta_modules = {}
		self.meta_workflow(self.meta_modules)

	def meta_workflow(self, meta_dict):
		'''Meta modules (analysis modules) are identified and executed

		Notes
		-----
		Naming convention for analysis module: in self.inp, the table title has to contain 
		`Analysis` and the last part of the string (seperated by spaces) has to be the module name'''

		for key in self.inp:
			if self.check_for_meta_module(key) is True:

				module_name = key.split(' ')[-1]
				self.execute_meta_module(module_name, meta_dict)

	def execute_meta_module(self, module_name, meta_dict):
		'''Requested module class is executed.
		'''
		
		module = execute_plugin(module_name, meta_dict, 
								plugin_module = False, 
								nested_dictionary = True, 
								input_file = self.input_file)

		for key in self.inp:
			if module_name in key and 'Methods' in key: 
				self.execute_module_methods(module, key, module_name, meta_dict)

	def execute_module_methods(self, module, key, module_name, meta_dict):
		'''Requested methods of module class are executed.
		'''

		for row_name, table in self.inp[key].items():
			method_name = table['Method Name']
			method = getattr(module, method_name)

			arguments = self.get_arguments(table)

			if self.check_for_plotting_method(method_name) is True:
				arguments['directory'] = self.output_directory

			meta_dict[module_name][row_name] = method(**arguments)

	def get_arguments(self, table):
		'''Arguments are read from the table in self.inp referenced in
		table['Arguments'] or directly read from table['Arguments']
		'''

		try:
			arguments = table['Arguments']
		except KeyError:
			return {}

		try:
			return(convert_dict_to_kwargs_dict(self.inp[arguments]))
		except KeyError:
			return(ast.literal_eval(arguments))

	def check_for_plotting_method(self, method_name):
		'''Returns true if a plotting indicator substring is in
		`method_name`.
		'''

		plotting_indicators = ['plot', 'figure', 'chart']

		if any(indicator in method_name for indicator in plotting_indicators):
			return True
		else:
			return False

	def check_for_meta_module(self, key):
		'''Checks if `key` is a meta module that is to be executed.

		Notes
		-----
		Meta module is identified by checking if `key` contains the substring
		'Analysis' and does not contain any of the substrings in `exceptions`.
		'''

		exceptions = ['Parameters', 'Methods', 'Arguments']
		indicators = ['Analysis']

		if any(exception in key for exception in exceptions):
			return False
		elif any(indicator in key for indicator in indicators):
			return True
		else:
			return False

def run_pyH2A():
	'''Wrapper function to run pyH2A from the command line.
	'''

	path_to_input = sys.argv[1]
	path_to_output = sys.argv[2] 

	output = pyH2A(path_to_input, path_to_output)

	return output

def command_line_pyH2A(input_file, output_dir):
	'''Wrapper function to run pyH2A using click.
	'''

	output = pyH2A(input_file, output_dir)

	return output




from input_modification import insert, process_table
import pprint


class Production_Scaling_Plugin:
	"""
	Calculation of plant output and potential scaling.

	If you want to scale capital or labor costs you have to specificy a path to `Scaling > Capital Scaling Factor > Value`
	or `Scaling > Labor Scaling Factor > Value` for the respective table entry.
	...

	Parameters
	----------
    test : test
	[...] Indirect Capital Costs [...] > > Value : float
		`sum_all_tables` is used.
	Technical Operating Parameters and Specifications > Plant Design Capacity (kg of H2/day) > Value : float
		Plant design capacity in kg of H2/day, `process_table()` is used.
	Technical Operating Parameters and Specifications > Operating Capacity Factor (%) > Value : float
		Operating capacity factor in %, `process_table()` is used.
	Technical Operating Parameters and Specifications > Maximum Output at Gate > Value : float, optional
		Maximum output at gate in kg of H2/day, 'process_table()' is used. If this parameter is
		not specified it defaults to `Plant Design Capacity (kg of H2/day)`.
	Technical Operating Parameters and Specifications > New Plant Design Capacity (kg of H2/day) > Value : float, optional
		New plant design capacity in kg of H2/day to calculate scaling, which overwrites possible Scaling Ratio,
		`process_table()` is used.
	Technical Operating Parameters and Specifications > Scaling Ratio > Value : float, optional
		Scaling ratio which is multiplied by current plant design capacity to obtain scaled plant size,
		`process_table` is used.
	Technical Operating Parameters and Specifications > Capital Scaling Exponent > Value : float, optional
		Exponent to calculate capital scaling factor, `process_table()` is used. Defaults to 0.78.
	Technical Operating Parameters and Specifications > Labor Scaling Exponent > Value : float, optional
		Exponent to calculcate labor scaling factor, `process_table()` is used. Defaults to 0.25.

	Returns
	-------
	Technical Operating Parameters and Specifications > Design Output per Day > Value : float
		Design output in kg of H2/day.
	Technical Operating Parameters and Specifications > Max Gate Output per Day > Value : float
		Maximum gate ouput in kg of H2/day.
	Technical Operating Parameters and Specifications > Output per Year > Value : float
		Yearly output taking operating capacity factor into account, in kg H2/year.
	Technical Operating Parameters and Specifications > Output per Year at Gate > Value	: float
		Actual yearly output at gate, in kg H2/year.
	Technical Operating Parameters and Specifications > Scaling Ratio > Value : float or None
		Returned if New Plant Design Capacity was specified.
	Scaling > Capital Scaling Factor > Value : float or None
		Returned if scaling is active (Scaling Ratio or New Plant Design Capacity specified).
	Scaling > Labor Scaling Factor > Value : float or None
		Returned if scaling is active (Scaling Ratio or New Plant Design Capacity specified).
	"""
	
	def __init__(prod, self, print_info):
		process_table(self.inp, 'Technical Operating Parameters and Specifications', 'Value')

		prod.dictionary = self.inp['Technical Operating Parameters and Specifications']

		prod.calculate_scaling(self, print_info)
		prod.calculate_output(self)

		insert(self, 'Technical Operating Parameters and Specifications', 'Design Output per Day', 'Value', prod.design_output_per_day, __name__, print_info = print_info)
		insert(self, 'Technical Operating Parameters and Specifications', 'Max Gate Output per Day', 'Value', prod.max_gate_output_per_day, __name__, print_info = print_info)

		insert(self, 'Technical Operating Parameters and Specifications', 'Output per Year', 'Value', prod.output_per_year, __name__, print_info = print_info)
		insert(self, 'Technical Operating Parameters and Specifications', 'Output per Year at Gate', 'Value', prod.output_per_year_at_gate, __name__, print_info = print_info)

	def calculate_scaling(prod, self, print_info):

		if 'Maximum Output at Gate' not in prod.dictionary:
			maximum_output_at_gate = prod.dictionary['Plant Design Capacity (kg of H2/day)']['Value']
			insert(self, 'Technical Operating Parameters and Specifications', 'Maximum Output at Gate', 'Value', maximum_output_at_gate, __name__, print_info = print_info)
	
		if 'New Plant Design Capacity (kg of H2/day)' in prod.dictionary:
			scaling_ratio = prod.dictionary['New Plant Design Capacity (kg of H2/day)']['Value'] / prod.dictionary['Plant Design Capacity (kg of H2/day)']['Value']
			insert(self, 'Technical Operating Parameters and Specifications', 'Scaling Ratio', 'Value', scaling_ratio, __name__, print_info = print_info)

		if 'Scaling Ratio' in prod.dictionary:
			prod.design_output_per_day = prod.dictionary['Plant Design Capacity (kg of H2/day)']['Value'] * prod.dictionary['Scaling Ratio']['Value']
			prod.max_gate_output_per_day = prod.dictionary['Maximum Output at Gate']['Value'] * prod.dictionary['Scaling Ratio']['Value']

			if 'Capital Scaling Exponent' in prod.dictionary:
				prod.capital_scaling_factor = prod.dictionary['Scaling Ratio']['Value'] ** prod.dictionary['Capital Scaling Exponent']['Value']
			else:
				prod.capital_scaling_factor = prod.dictionary['Scaling Ratio']['Value'] ** 0.78

			if 'Labor Scaling Exponent' in prod.dictionary:
				prod.labor_scaling_factor = prod.dictionary['Scaling Ratio']['Value'] ** prod.dictionary['Labor Scaling Exponent']['Value']
			else:
				prod.labor_scaling_factor = prod.dictionary['Scaling Ratio']['Value'] ** 0.25

			insert(self, 'Scaling', 'Capital Scaling Factor', 'Value', prod.capital_scaling_factor, __name__, print_info = print_info)
			insert(self, 'Scaling', 'Labor Scaling Factor', 'Value', prod.labor_scaling_factor, __name__, print_info = print_info)

		else:
			prod.design_output_per_day = prod.dictionary['Plant Design Capacity (kg of H2/day)']['Value']
			prod.max_gate_output_per_day = prod.dictionary['Maximum Output at Gate']['Value']

	def calculate_output(prod, self):

		prod.output_per_year = prod.design_output_per_day * 365. * prod.dictionary['Operating Capacity Factor (%)']['Value']
		prod.output_per_year_at_gate = prod.max_gate_output_per_day * 365. * prod.dictionary['Operating Capacity Factor (%)']['Value']



def detect_parameters_and_output(line):
	""""Detection of parameters and output values in line based on presence of more than one tab
	or more than four spaces"""

	if '\t\t' in line: #or '     ' in line:
		return False
	else:
		return True

def process_single_line(line, output_dict):

	if detect_parameters_and_output(line):

		complete_string = line.split(':')
		variable_string = complete_string[0].strip(', \t')
		variable_type = complete_string[1].strip(', \t')

		output_dict[variable_string] = variable_type

def extract_input_output_from_docstring(plugin_name):

	doc_string = plugin_name.__doc__.split('\n')

	parameters_dict = {}
	output_dict = {}

	parameters = False
	output = False

	for counter, line in enumerate(doc_string):

		if '---' in line and 'Parameters' in doc_string[counter-1]:
			parameters = True

		if '---' in line and 'Returns' in doc_string[counter-1]:
			parameters = False
			output = True
	
		if line == '' or line == '\t':
			parameters = False
			output = False

		if parameters and '---' not in line:
			process_single_line(line, parameters_dict)

		if output and '---' not in line:
			process_single_line(line, output_dict)


	print(output_dict)




extract_input_output_from_docstring(Production_Scaling_Plugin)
		


from pyH2A.Utilities.Energy_Conversion import Energy, kWh, eV
from pyH2A.Utilities.input_modification import insert, process_table

class Solar_Thermal_Plugin:
	'''Simulation of hydrogen production using solar thermal water splitting.

	Parameters
	----------
	Technical Operating Parameters and Specifications > Design Output per Day > Value : float
		Design output of hydrogen production plant per day in kg.
	Solar-to-Hydrogen Efficiency > STH (%) > Value : float
		Solar-to-Hydrogen Efficiency of thermal water splitting process. Percentage of value 
		between 0 and 1.
	Solar Input > Mean solar input (kWh/m2/day) > Value : float
		Mean solar input in kWh/m2/day.
	Non-Depreciable Capital Costs > Additional Land Area (%) > Value : float
		Additional land area required. Percentage or value > 0. Calculated as:
		(1 + Addtional Land Area) * solar collection area.

	Returns
	-------
	Non-Depreciable Capital Costs > Land required (acres) > Value : float
		Total land requirement in acres.
	'''
	
	def __init__(therm, dcf, print_info):
		process_table(dcf.inp, 'Technical Operating Parameters and Specifications', 'Value')
		process_table(dcf.inp, 'Solar-to-Hydrogen Efficiency', 'Value')
		process_table(dcf.inp, 'Solar Input', 'Value')
		process_table(dcf.inp, 'Non-Depreciable Capital Costs', 'Value')

		therm.calculate_land_area(dcf)

		insert(dcf, 'Non-Depreciable Capital Costs', 'Land required (acres)', 'Value', therm.area_acres, __name__, print_info = print_info)

	def calculate_land_area(therm, dcf):
		'''Calculation of required land area based on solar input, solar-to-hydrogen efficiency
		and addtional land are requirements.
		'''

		insolation_per_m2_per_day = Energy(dcf.inp['Solar Input']['Mean solar input (kWh/m2/day)']['Value'], kWh)

		mol_H2_per_m2_per_day = (insolation_per_m2_per_day.J * dcf.inp['Solar-to-Hydrogen Efficiency']['STH (%)']['Value']) / Energy(2*1.229, eV).Jmol
		kg_H2_per_m2_per_day = (2 * mol_H2_per_m2_per_day)/1000.

		required_area_m2 = dcf.inp['Technical Operating Parameters and Specifications']['Design Output per Day']['Value'] / kg_H2_per_m2_per_day

		therm.area_m2 = required_area_m2 * (1. + dcf.inp['Non-Depreciable Capital Costs']['Additional Land Area (%)']['Value'])
		therm.area_acres = therm.area_m2 * 0.000247105



import multiprocessing
import copy
from pathlib import Path
from timeit import default_timer as timer
import numpy as np
from scipy.signal import savgol_filter
from scipy.spatial import distance as scipy_distance
from scipy.stats import norm as normal_distribution
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
from matplotlib.transforms import Bbox

import pyH2A.Utilities.find_nearest as fn
from pyH2A.Utilities.input_modification import convert_input_to_dictionary,parse_parameter, parse_parameter_to_array, get_by_path, set_by_path, read_textfile, file_import
from pyH2A.Discounted_Cash_Flow import Discounted_Cash_Flow
from pyH2A.Utilities.output_utilities import make_bold, format_scientific, dynamic_value_formatting, insert_image, Figure_Lean

def select_non_reference_value(reference, values):
	'''Select value from values which is not the reference one.
	'''

	idx = np.invert(np.equal(values, reference))
	return values[idx][0]

def divide_into_batches(array, batch_size):
	'''Divide provided array into batches of size batch_size for parallel processing'''

	number_of_divisions = np.floor(len(array)/batch_size)
	idx = int(number_of_divisions * batch_size)

	first_part = array[:idx]
	second_part = array[idx:]

	batches = np.split(first_part, number_of_divisions)

	if second_part.size != 0:
		batches.append(second_part)

	return batches

def normalize_parameter(parameter, range_arr):
	'''Linear normalization of parameter (float or array) based on 
	range_arr (lower limit: range_arr[0],
	upper limit: range_arr[1])
	'''

	range_covered = range_arr[1] - range_arr[0]
	scaled = (parameter - range_arr[0]) / range_covered

	return scaled

def extend_limits(limits_original, extension):
	'''Extend limits_original in both directions by muliplying with extensions'''

	limits = np.copy(limits_original)

	limit_range = limits[1] - limits[0]

	limits[0] -= extension * limit_range
	limits[1] += extension * limit_range

	return limits

def coordinate_position(x_reference, x_values, y_reference, y_values):
	'''Determine correct position for base case label in `plot_colored_scatter()`.
	'''

	if x_reference - select_non_reference_value(x_reference, x_values) > 0:
		xtext = 0.85
	else:
		xtext = 0.05

	if y_reference - select_non_reference_value(y_reference, y_values) > 0:
		ytext = 0.93
	else:
		ytext = 0.1

	return xtext, ytext

class Monte_Carlo_Analysis:
	'''Monte Carlo analysis of a techno-economic model.

	Parameters 
	----------
	Monte_Carlo_Analysis > Samples > Value : int
		Number of samples for Monte Carlo analysis.
	Monte_Carlo_Analysis > Target Price Range ($) > Value : str
		Target price range for H2 specfied in the following format:
		lower value; higher value (e.g. "1.5: 1.54").
	Monte_Carlo_Analysis > Output File > Value : str, optional
		Path to location where output file containing Monte Carlo analysis
		results should be saved.
	Monte_Carlo_Analysis > Input File > Value : str, optional
		Path to location of file containing Monte Carlo analysis results that
		should be read.
	Parameters - Monte_Carlo_Analysis > [...] > Name : str
		Display name for parameter, e.g. used for axis labels.
	Parameters - Monte_Carlo_Analysis > [...] > Type : str
		Type of parameter values. If `Type` is 'value', provided values are
		used as is. If `Type` is 'factor', provided values are multiplied 
		with base value of parameter in input file.
	Parameters - Monte_Carlo_Analysis > [...] > Values : str
		Value range for parameter in Monte Carlo analysis. Specified in following
		format: upper limit; lower limit (order is irrelevant). Instead of actual values,
		'Base' or 'Reference' can be used to retrieve base value of parameter in input file
		as one of the values. E.g. 'Base; 20%'.
	Parameters - Monte_Carlo_Analysis > [...] > File Index : int, optional
		If Monte Carlo results are read from a provided input file, `File Index` for each
		parameter can be specified. `File Index` refers to the column position of each parameter
		in the read input file. This mapping allows for changing the display name and position
		of a parameter in the `Parameters - Monte_Carlo_Analysis` analysis table and still
		ensure correct mapping of the read results.

	Notes
	-----
	`Parameters - Monte_Carlo_Analysis` contains parameters which are to be varied in
	Monte Carlo Analysis. First column specifies path to parameter in input file 
	(top key > middle key > bottom key format, e.g. Catalyst > Cost per kg ($) > Value).
	Order of parameters can be changed, which for example affects the mapping onto different
	axis in `plot_colored_scatter` (first parameter is on x axis, second on y axis, etc.).
	'''

	def __init__(self, input_file):
		'''Initialization of Monte Carlo analysis.

		Notes 
		-----
		`self.inp` is generated from provided `input_file`. 
		If 'Input File' is specified in Monte_Carlo_Analysis table of`input_file`, 
		the provided file (which has to contain Monte Carlo simulation 
		data formatted as done by `save_results()` method) is read and Monte Carlo simulation
		results and corresponding parameter specification are retrieved.
		If 'Input File' is not specified, the parameters specified in `input_file` are processed 
		and used to generate Monte Carlo simulation data (using `perform_monte_carlo_multiprocessing()`), 
		which are stored in the specified 'Output File'.
		Subsequently, Monte Carlo datapoints are selected based on the specified target price range, 
		development distances are calculated and plots can be generated.
		'''

		self.inp = convert_input_to_dictionary(input_file)

		if 'Display Parameters' in self.inp:
			self.color = self.inp['Display Parameters']['Color']['Value']
			self.display_name = self.inp['Display Parameters']['Name']['Value']
		else:
			self.color = 'darkgreen'
			self.display_name = 'Model'

		if 'Input File' in self.inp['Monte_Carlo_Analysis']:
			self.read_results(self.inp['Monte_Carlo_Analysis']['Input File']['Value'])
		else:
			self.process_parameters()
			self.perform_full_monte_carlo()
			self.save_results(self.inp['Monte_Carlo_Analysis']['Output File']['Value'])

		self.check_parameter_integrity(self.results)
		self.target_price_components()
		self.determine_principal_components()
		self.generate_parameter_string_table()
		self.development_distance()
		self.full_distance_cost_relationship()

	def process_parameters(self):
		'''
		Monte Carlo Analysis parameters are read from 'Monte Carlo Analysis - Parameters' 
		in `self.inp` and processed.

		Notes
		-----
		The ranges for each parameter are defined in `Values` column as ';' seperated entries. Entries can
		either be a number, a path, or a `special_value` such as `Base` or `Reference`. If such a `special_value`
		is specified, the base value of that parameter is retrieved from `self.inp`.
		Parameter information is stored in `self.parameters` attribute.
		Based on the ranges for each parameter, random values (uniform distribution) are generated and stored
		in the `self.values` attribute.
		The target price range is read from `self.inp` file and stored in `self.target_price_range` attribute.
		'''

		monte = self.inp['Parameters - Monte_Carlo_Analysis']
		samples = self.inp['Monte_Carlo_Analysis']['Samples']['Value']

		number_parameters = len(monte)

		values = np.empty((samples, number_parameters))
		parameters = {}

		for counter, key in enumerate(monte):
			values_range = parse_parameter_to_array(monte[key]['Values'], delimiter = ';', 
													dictionary = self.inp, 
													top_key = 'Parameters - Monte_Carlo_Analysis', 
													middle_key = key, bottom_key = 'Values', 
													special_values = ['Base', 'Reference'], 
													path = key)

			values_range = values_range[np.argsort(values_range)]
			values[:,counter] = np.random.uniform(values_range[0], 
												  values_range[1], 
												  samples)

			path = parse_parameter(key)
			reference = get_by_path(self.inp, path)
		
			parameters[monte[key]['Name']] = {'Parameter': path, 'Type': monte[key]['Type'], 
											  'Values': values_range, 'Reference': reference, 
											  'Index': counter, 'Input Index': counter}
	
		self.values = values
		self.parameters = parameters
		self.target_price_range = parse_parameter_to_array(self.inp['Monte_Carlo_Analysis']['Target Price Range ($)']['Value'], 
														   delimiter = ';', 
														   dictionary = self.inp)

	def perform_h2_cost_calculation(self, values):
		'''H2 cost calculation for provided parameter values is performed.

		Parameters
		----------
		values : ndarray
			Array containing parameter variations.

		Returns
		-------
		h2_cost : ndarray
			1D array of H2 cost values for each set of parameters.

		Notes
		-----
		Performs H2 cost calulation by modifying a copy of self.inp based 
		on the provided values and `self.parameters`. The modified copy of 
		`self.inp` is then passed to `Discounted_Cash_Flow()`.
		A parameter value can be either a value replacing the existing one 
		in self.inp (Type = value) or it can be a factor hich will be multiplied 
		by the existing value.
		'''

		parameters = self.parameters
		h2_cost = []

		for value_set in values:
			input_dict = copy.deepcopy(self.inp)

			for key, parameter in parameters.items():
				set_by_path(input_dict, parameter['Parameter'], value_set[parameter['Index']], 
							value_type = parameter['Type'])

			dcf = Discounted_Cash_Flow(input_dict, print_info = False)
			h2_cost.append(dcf.h2_cost)

		return np.asarray(h2_cost)

	def perform_monte_carlo_multiprocessing(self, values, return_full_array = True):
		'''Monte Carlo analysis is performed with multiprocessing parallelization across
		all available CPUs.

		Parameters
		----------
		values : ndarray
			2D array containing parameters variations which are to be evaluated.
		return_full_array : bool, optional
			If `return_full_array` is True, the full 2D array containing parameter
			variations and H2 cost is returned. Otherwise, a 1D array containing only
			H2 cost values is returned.

		Returns
		-------
		full_array : ndarray
			2D array containing parameter variations and H2 cost values.
		h2_cost : ndarray
			1D array containing H2 costvalues.
		'''

		num_cpus = multiprocessing.cpu_count()
		pool = multiprocessing.Pool(num_cpus)

		value_batches = divide_into_batches(values, np.ceil(len(values)/num_cpus))
		
		h2_cost = pool.map(self.perform_h2_cost_calculation, value_batches)
		h2_cost = np.concatenate(h2_cost)

		if return_full_array is True:
			return np.c_[self.values, h2_cost]
		else:
			return h2_cost

	def perform_full_monte_carlo(self):
		'''Monte Carlo analysis is performed based on random parameter variations 
		in `self.values`.
		'''

		start = timer()

		self.results = self.perform_monte_carlo_multiprocessing(self.values)

		end = timer()
		print('Time Monte Carlo Multi:', end - start)

	def save_results(self, file_name):
		'''Results of Monte Carlo simulation are saved in `file_name` and a 
		formatted header is added. Contains name, parameter path, type and values range 
		from `self.parameters`.
		'''

		header_string = ''
		path_string = ''
		type_string = ''
		values_string = ''

		for key in self.parameters:
			header_string += str(key) + '	'
			path_string += str(self.parameters[key]['Parameter']) + '	'
			type_string += str(self.parameters[key]['Type']) + '	'
			values_string += str(self.parameters[key]['Values']) + '	'

		header_string += 'H2 Cost'
		complete_string = header_string + '\n' + path_string + '\n' + type_string + '\n' + values_string

		np.savetxt(Path(file_name), self.results, header = complete_string, delimiter = '	')

	def read_results(self, file_name):
		'''Reads Monte Carlo simulation results from `file_name`.

		Parameters
		----------
		file_name : str
			Path to file containing Monte Carlo simulation results.	

		Returns
		-------
		self.results : ndarray
			Array containing parameters and H2 cost for each model.
		self.parameters : dict
			Dictionary containing information on varied parameters.
		self.target_price_range : ndarray
			Selected target price range from `self.inp`.

		Notes
		-----
		Assumes formatting created by `self.save_results()`function.
		Header must contain name of parameters, path to parameters in input file, 
		type of parameter and value range.
		The header is processed to retrieve these atrribtues and stores them in `self.parameters`.
		Reference values for each parameter and target price range are read from self.inp
		The order of parameters is also read from `self.inp` and stored in `self.parameters` 
		as `Input Index`. If the name of a parameter has been changed in `self.inp` and is 
		different from the parameter stored in `file_name`, it is checked whether a `File Index`
		is specified, which allows for mapping of renamed parameter to parameter stored in 
		`File Index`. If `File Index` is specified, the existing parameter at this position 
		in `File Index` is renamed to the specified name.
		'''

		self.results = read_textfile(file_name, delimiter = '	', mode = 'r')

		parameters = {}
		column_dict = {}
		row_dict = {0: 'Key', 1: 'Parameter', 2: 'Type', 3: 'Values'}

		file_read = file_import(file_name, mode = 'r')

		for row_counter, line in enumerate(file_read):

			if line[0] != '#':
				break
			else:
				line_clean = line.strip(' #\n')
				line_split = parse_parameter(line_clean, delimiter = '	')

				for column_counter, element in enumerate(line_split):

					if row_dict[row_counter] == 'Key':
						parameters[element] = {}
						parameters[element]['Index'] = column_counter
						column_dict[column_counter] = element

					else:
						row_key = row_dict[row_counter]
						column_key = column_dict[column_counter]
						element = element.strip('[]')

						if row_key == 'Parameter':
							arr = []
							split = element.split(',')
							for i in split:
								arr.append(i.strip("' "))
							element = arr
				
						if row_key == 'Values':
							element = np.fromstring(element, sep = ' ', dtype = float)

						parameters[column_key][row_key] = element

		file_read.close()

		del parameters['H2 Cost']

		for key in parameters:
			parameters[key]['Reference'] = get_by_path(self.inp, parameters[key]['Parameter'])
		
		self.parameters = parameters
		self.target_price_range = parse_parameter_to_array(self.inp['Monte_Carlo_Analysis']['Target Price Range ($)']['Value'], delimiter = ';')

		for counter, (key, parameter) in enumerate(self.inp['Parameters - Monte_Carlo_Analysis'].items()):
			
			if parameter['Name'] in self.parameters:
				self.parameters[parameter['Name']]['Input Index'] = counter # Identical name, storing Input Index

			elif 'File Index' in parameter:   # Non-identical name, looking for File Index to map
				for self_key, self_parameter in self.parameters.items():

					sucess = False

					if self_parameter['Index'] == parameter['File Index']:
						self_parameter['Input Index'] = counter
						self.parameters[parameter['Name']] = self.parameters.pop(self_key)
						sucess = True

				if sucess is False:  # Raise error it not all parameters with different names could be mapped
					raise KeyError('Input Parameter {0} could not be mapped.'.format(key))

			else:
				raise KeyError('Difference between input parameter names and those stored in Monte Carlo File, no index for mapping specified.')

	def check_parameter_integrity(self, values):
		'''Checking that parameters in `self.results` are within 
		ranges specified in `self.parameters['Values']`.
		'''
		for name, parameter in self.parameters.items():
			results = values[:,parameter['Index']]
			
			minimum = np.amin(results)
			maximum = np.amax(results)

			assert minimum >= parameter['Values'][0], 'Minimum value of {0} ({1}) is smaller than specified range ({2}).'.format(name, minimum, parameter['Values'])
			assert maximum <= parameter['Values'][1], 'Maximum value of {0} ({1}) is larger than specified range ({2}).'.format(name, maximum, parameter['Values']) 

	def generate_parameter_string_table(self, base_string = 'Base', limit_string = 'Limit'):
		'''String of parameter table is generated, used in `self.render_parameter_table`.

		Parameters 
		----------
		base_string : str, optional
			String used to label base column.
		limit_string : str, optinal
			String used to label limit column.
		'''

		self.parameter_string = make_bold(self.display_name) + '\n'
		self.parameter_table = {'Table Data': [[make_bold(self.display_name), 
												make_bold(base_string), '', 
												make_bold(limit_string)]], 'Row Labels': []}

		for key, value in self.parameters.items():
			reference_value = dynamic_value_formatting(value['Reference'])
			limit_value = dynamic_value_formatting(select_non_reference_value(value['Reference'], 
																			  value['Values']))

			string = key + ': ' + reference_value + r'$\rightarrow$' + limit_value + '\n'
			self.parameter_string += string

			self.parameter_table['Table Data'].append([key, reference_value, 
													   r'$\longrightarrow$', limit_value])
			self.parameter_table['Row Labels'].append(key)

	def render_parameter_table(self, ax, xpos = 1.05, ypos = 0.0, height = 1.0, 
							   colWidths = [0.55, 0.25, 0.07, 0.25],
							   left_pad = 0.01, edge_padding = 0.02, 
							   fontsize = 12, base_string = 'Base',
							   limit_string = 'Limit'):
		'''Rendering table of parameters which are varied during Monte Carlo analysis.

		Parameters
		----------
		ax : matplotlib.axes
			Axes object in which parameter table is displayed.
		xpos : float, optional
			x axis position of left edge of table in axis fraction coordinates.
		ypos : float, optional
			y axis position of lower edge of table in axis fraction coordinates.
		height : float, optional
			Height of table in axis fraction coordinates (e.g. height = 0.5 meaning that
			the table has half the height of the plot).
		colWidths : list, optional
			List of length 4 with widths for each column from left to right in axis
			fraction coordinates.
		left_pad : float, optional
			Padding of the table on the left site.
		edge_padding: float, optional
			Padding of edges that are drawn.
		fontsize : float, optional
			fontsize for table.
		base_string : str, optional
			String used to label base column.
		limit_string : str, optional
			String used to label limit column.

		Notes
		-----
		Table is rendered in provided matplotlib.axes object.
		'''

		self.generate_parameter_string_table(base_string = base_string)

		bbox = [xpos, ypos, sum(colWidths), height]
		number_of_rows = len(self.parameter_table['Table Data'])
		cell_height = bbox[3] / number_of_rows

		table = ax.table(cellText = self.parameter_table['Table Data'],
						 edges = 'open',
						 bbox = bbox,
						 colWidths = colWidths,
						 cellLoc = 'left')

		for key, cell in table.get_celld().items():
			cell.get_text().set_color('black')
			if key[0] == 0 or key[1] >= 1:
				cell.get_text().set_color(self.color)

		table.auto_set_font_size(False)
		table.set_fontsize(fontsize)

		def set_pad_for_column(col, pad):
			cells = table.get_celld()
			column = [cell for cell in table.get_celld() if cell[1] == col]
			for cell in column:
				cells[cell].PAD = pad

		set_pad_for_column(col=0, pad = left_pad)

		def plot_edge(row_position, edge_padding):

			if row_position == 'last':
				y_edge = bbox[1] + edge_padding
			else:
				y_edge = bbox[1] + bbox[3] - (row_position * cell_height) + edge_padding

			table_edge = ax.annotate('', xy = (bbox[0], y_edge), 
									 xytext = (bbox[0] + bbox[2], y_edge), 
									 xycoords = ax.transAxes, textcoords = ax.transAxes, 
									 arrowprops={'arrowstyle': '-', 'color': self.color})

		plot_edge(1, edge_padding)
		plot_edge('last', 0)

	def target_price_components(self):
		'''Monte Carlo simulation results are sorted by H2 cost and the entries of 
		`self.results` with a H2 cost within the specified target price range are stored 
		in `self.target_price_data`.
		'''

		results_sorted = self.results[np.argsort(self.results[:,-1])]
		idx = fn.find_nearest(results_sorted[:,-1], self.target_price_range)
		data = results_sorted[idx[0]:idx[1]]

		self.target_price_data = data

	def determine_principal_components(self):
		'''Converting parameters to list sorted by input index.
		'''

		# placeholder for actual PCA
		# parameter/principal component class instead of dict
		# map PCs to W, X, Y, Z?

		self.principal = [None] * len(self.parameters)
		self.base_case = [None] * len(self.parameters)

		for key, parameter in self.parameters.items():
			self.principal[parameter['Input Index']] = key
			self.base_case[parameter['Index']] = parameter['Reference']

	def target_price_2D_region(self, grid_points = 15):
		'''Determining largest region spanned by first two parameters within which
		target prices can be achieved.

		Parameters
		----------
		grid_points : int, optional
			Number of grid points to determine density of grid evaluation.

		Returns
		-------
		self.target_price_2D_region : dict
			Dict of ndarrays with information of target price 2D region.

		Notes
		-----		
		Model is evaluated on grid spanned by first two parameters (density of grid is controlled by 
		grid_points), other parameters are set to limit (non-reference) values.
		Output is a dictionary (`self.target_price_2D_region`), which can be used to overlay 
		target price region onto scatter ploting using plt.contourf
		'''

		grid_axes = np.empty((2, grid_points))
		grid_idx = []

		values = np.empty((grid_points**2, len(self.parameters)))

		for key, parameter in self.parameters.items():
			if parameter['Input Index'] < 2:
				value_range = parameter['Values']

				grid_axis = np.linspace(value_range[0], value_range[1], grid_points)
				grid_axes[parameter['Input Index']] = grid_axis
				grid_idx.append([parameter['Index'], parameter['Input Index']])

			else:
				used_value = select_non_reference_value(parameter['Reference'], parameter['Values'])
				values[:,parameter['Index']] = np.ones(len(values)) * used_value

				parameter['Target Price Range'] = {}
				parameter['Target Price Range']['Range'] = self.target_price_range
				parameter['Target Price Range']['Used Value'] = used_value

		grid_values = np.meshgrid(*grid_axes)
		grid_values_ravel = np.c_[[np.ravel(i) for i in grid_values]].T

		for idx, idx_input in grid_idx:
			values[:,idx] = grid_values_ravel[:,idx_input]

		self.check_parameter_integrity(values)	

		h2_cost = self.perform_monte_carlo_multiprocessing(values, return_full_array = False)
		h2_cost_2D = np.reshape(h2_cost, (grid_points, grid_points))

		self.target_price_2D_region = {'Grid Values': grid_values, 'H2 Cost 2D': h2_cost_2D}

	def calculate_distance(self, data, parameters, selection, metric):
		'''
		Distance of datapoints to reference is calculated using the specified metric.

		Parameters
		----------
		data: ndarray
			2D array of datapoints containing parameter values for each model.
		parameters: dict
			Dictionary specifying ranges for each parameter.
		selection: list
			List of parameters names used for distance calculation.

		Returns
		-------
		distances: ndarray
			Array containing distance for each model.

		Notes
		-----
		Parameter ranges and the reference parameters are scaled to be within 
		a n-dimensional unit cube.
		Distances are normalized by the number of dimensions, so that the maximum 
		distance is always 1.
		'''

		number_of_parameters = len(selection)

		reference = []
		data_scaled = np.empty((len(data), number_of_parameters))

		for counter, key in enumerate(selection):
			range_arr = parameters[key]['Values']

			idx = parameters[key]['Index']
			data_scaled[:,counter] = normalize_parameter(data[:,idx], range_arr)
			reference.append(normalize_parameter(parameters[key]['Reference'], range_arr))

		reference_scaled = np.array([reference])

		distances = scipy_distance.cdist(data_scaled, reference_scaled, metric = metric)

		if metric is 'cityblock':
			distances = distances / number_of_parameters
		else:
			distances = distances / np.sqrt(number_of_parameters)

		return distances

	def development_distance(self, metric = 'cityblock'):
		'''Calculation of development distance for models within target price range.

		Parameters
		----------
		metric: str, optional
			Metric used for distance calculation, defaults to `cityblock`.

		Returns
		-------
		self.distances: ndarrray
			Array containing distances for models within target price range.

		Notes
		-----
		The euclidean or cityblock distance in n-dimensional space of each Monte Carlo simulation datapoint within 
		the target price range to the reference point is calculated and stored in self.distances.
		Parameter ranges and the reference parameters are scaled to be within a n-dimensional unit cube.
		Distances are normalized by the number of dimensions, so that the maximum distance is always 1.'''

		self.distances = self.calculate_distance(self.target_price_data, self.parameters, self.principal, metric = metric)

	def full_distance_cost_relationship(self, metric = 'cityblock', reduction_factor = 25, 
										poly_order = 4):
		'''Calculation of development distance for all datapoints from Monte Carlo Analysis and
		calculation of Savitzky-Golay filter.

		Parameters
		----------
		metric : str, optional
			Distance metric used for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.calculate_distance`
		reduction_factor : int, optional
			Determines window size for Savitzky-Golay filter.
		poly_order : int, optional
			Order of polynomial for Savitzky-Golay filter

		Returns
		-------
		self.results_distances_sorted : ndarray
			Sorted array of distances for all datapoints from Monte Carlo Analysis.
		self.distances_cost_savgol : ndarray
			Savitzky-Golay filter results.
		'''

		window_length = int(len(self.results)/reduction_factor)

		if window_length % 2 == 0:
			window_length += 1

		distances = self.calculate_distance(self.results, self.parameters, self.principal, metric = metric)

		results_distances = np.c_[self.results, distances]
		self.results_distances_sorted = results_distances[np.argsort(results_distances[:,-1])]

		smoothed = savgol_filter(self.results_distances_sorted[:,-2], window_length, poly_order)
	
		self.distances_cost_savgol = np.c_[self.results_distances_sorted[:,-1], smoothed]

	def plot_complete_histogram(self, bins = None, xlim_low = None, xlim_high = None,
								xlabel_string = 'Levelized $H_{2}$ Cost / \$/kg', 
								ylabel_string = 'Normalized Frequency',
								image_kwargs = {}, plot_kwargs = {},
								**kwargs):
		'''Complete histogram of price distribution from Monte Carlo analysis

		Parameters 
		----------
		bins : int, optional
			Number of bins for histogram. If `None`, bins is calculated
			from the size of `self.results`.
		xlim_low : float or None, optional
			Lower x axis limit.
		xlim_high : float or None, optional
			Higher x axis limit.
		xlabel_string : str, optional
			String for x axis label.
		ylabel_string : str, optional
			String for y axis label.
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned.
		'''

		kwargs = {**{'right': 0.95, 'bottom': 0.15, 'top': 0.95,
	 			     'fig_width': 7, 'fig_height': 4, 'font_size': 12,
	 			     'name': 'Monte_Carlo_Complete_Histogram'}, 
	 			  **kwargs, **plot_kwargs}

		image_kwargs = {**{'path': None, 'x': 0.5, 'y': 0.8, 'zoom': 0.08}, 
						**image_kwargs}

		figure = Figure_Lean(**kwargs)
		ax = figure.ax

		if bins is None:
			bins = int(len(self.results) / 20)

		ax.hist(self.results[:,-1], bins = bins, density=True, color=self.color, edgecolor = 'black')

		ax.set_xlabel(xlabel_string)
		ax.set_ylabel(ylabel_string)

		if xlim_low is not None:
			ax.set_xlim(xlim_low, xlim_high)

		if image_kwargs['path'] is not None:
			insert_image(ax = ax, **image_kwargs)

		figure.execute()

		return figure.fig

	def plot_colored_scatter(self, limit_extension = 0.03,   
							 title_string = 'Target Price Range: ', 
							 base_string = 'Base',
							 image_kwargs = {}, plot_kwargs = {},  
							 **kwargs):
		'''Plotting colored scatter plot showing all models within target price range.

		Parameters
		----------
		limit_extension: float, optional
			Amount of limit extension of axes as fraction of axis range.
		title_string : str, optional
			String for title.
		base_string : str, optional
			String to label base case datapoint.
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned.

		Notes
		-----
		x, y and color axis are determined by 
		:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.determine_principal_components`,
		with pc[0] being the x axis, pc[1] the y axis and pc[2] the color axis. Order can
		be changed by changing order of parameters in input file.
		'''

		kwargs = {**{'left': 0.15, 'right': 0.9, 'bottom': 0.11, 'top': 0.88,
	 			     'fig_width': 6.4, 'fig_height': 4.8, 'font_size': 12,
	 			     'name': 'Monte_Carlo_Colored_Scatter'}, 
	 			  **kwargs, **plot_kwargs}

		image_kwargs = {**{'path': None, 'x': 0.5, 'y': 0.8, 'zoom': 0.08}, 
						**image_kwargs}

		pc = self.principal
		par = self.parameters

		figure = Figure_Lean(**kwargs)
		ax = figure.ax
		fig = figure.fig

		cm = plt.get_cmap('plasma')

		self.target_price_2D_region()

		contour_fill = ax.contourf(*self.target_price_2D_region['Grid Values'], 
							        self.target_price_2D_region['H2 Cost 2D'],
							        levels = [0., max(self.target_price_range)], 
							        		  alpha = 0.1, colors = [cm(0.0)])

		base_case_h2_cost_appended = np.r_[self.base_case, 0]
		target_price_data_appended = np.vstack((self.target_price_data, 
												base_case_h2_cost_appended))

		scatter = ax.scatter(target_price_data_appended[:,par[pc[0]]['Index']], 
							 target_price_data_appended[:,par[pc[1]]['Index']], 
						 c = target_price_data_appended[:,par[pc[2]]['Index']], 
						 	 cmap = 'plasma', alpha = 1.)

		colorbar = fig.colorbar(scatter, ax=ax)

		ax.set_xlim(extend_limits(par[pc[0]]['Values'], limit_extension))
		ax.set_ylim(extend_limits(par[pc[1]]['Values'], limit_extension))
		scatter.set_clim(par[pc[2]]['Values'])

		ax.set_xlabel(pc[0])
		ax.set_ylabel(pc[1])
		colorbar.set_label(pc[2])

		xtext, ytext = coordinate_position(par[pc[0]]['Reference'], par[pc[0]]['Values'],
										   par[pc[1]]['Reference'], par[pc[1]]['Values']) 

		ax.annotate(make_bold(base_string), xy = (par[pc[0]]['Reference'], 
					par[pc[1]]['Reference']), xytext = (xtext, ytext), 
					textcoords = 'axes fraction')

		ax.set_title(make_bold(title_string) + '{0} - {1} \$/kg'.format(
											self.target_price_range[0], 
											self.target_price_range[1]))

		ax.grid(color = 'grey', linestyle = '--', linewidth = 0.2, zorder = 0)

		if image_kwargs['path'] is not None:
			insert_image(ax = ax, **image_kwargs)

		figure.execute()

		return figure.fig

	def plot_colored_scatter_3D(self, limit_extension = 0.03,   
							 	title_string = 'Target Price Range: ',
							 	**kwargs):
		'''3D colored scatter plot of models within target price range.
	
		Parameters
		----------
		limit_extension: float, optional
			Amount of limit extension of axes as fraction of axis range.
		title_string: str, optional
			Title string.

		Returns
		-------
		fig: matplotlib.figure
			Figure object.
		'''
		from mpl_toolkits.mplot3d import Axes3D
	
		pc = self.principal
		par = self.parameters

		fig = plt.figure()
		ax = Axes3D(fig, auto_add_to_figure=False)
		fig.add_axes(ax)

		scatter = ax.scatter(self.target_price_data[:,par[pc[0]]['Index']], 
							 self.target_price_data[:,par[pc[1]]['Index']],  
							 self.target_price_data[:,par[pc[3]]['Index']], 
							 c = self.target_price_data[:,par[pc[2]]['Index']], 
							 cmap = 'plasma', depthshade = False)
	
		colorbar = fig.colorbar(scatter, ax=ax, shrink = 0.9)

		ax.set_xlim(extend_limits(par[pc[0]]['Values'], limit_extension))
		ax.set_ylim(extend_limits(par[pc[1]]['Values'], limit_extension))
		ax.set_zlim(extend_limits(par[pc[3]]['Values'], limit_extension))
		scatter.set_clim(par[pc[2]]['Values'])

		ax.set_xlabel(pc[0])
		ax.set_ylabel(pc[1])
		ax.set_zlabel(pc[3])
		colorbar.set_label(pc[2])

		ax.set_title(make_bold(title_string) + '{0} - {1} \$/kg'.format(
											self.target_price_range[0], 
											self.target_price_range[1]))
	
		ax.grid(color = 'grey', linestyle = '--', linewidth = 0.2, zorder = 0)

		plt.show()

		return fig

	def plot_distance_histogram(self, ax = None, bins = 25, 
								figure_lean = True, xlabel = False, title = True,
								xlabel_string = 'Normalized Distance',
								ylabel_string = 'Frequency',
								title_string = 'Target Price Range',
								table_kwargs = {}, image_kwargs = {}, plot_kwargs = {},
								 **kwargs):

		'''Plotting development distances as histogram.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		bins : int, optional
			Number of bins for histogram.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		xlabel : bool, optional
			Flag to control if x axis label is displayed or not.
		title : bool, optional
			Flag to control if title is displayed or not.
		x_label_string : str, optional
			String for x axis label.
		y_label_string : str, optional
			String for y axis label.
		table_kwargs : dict, optional
			Dictionary containing optional keyword arguments for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.render_parameter_table`
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.

		'''

		kwargs = {**{'left': 0.2, 'right': 0.55, 'bottom': 0.25, 'top': 0.85, 'hspace': 0.2,
	 			     'fig_width': 11, 'fig_height': 2.5, 'font_size': 12,
	 			     'name': 'Monte_Carlo_Distance_Histogram'}, 
	 			  **kwargs, **plot_kwargs}

		table_kwargs = {**{'colWidths': [0.55, 0.25, 0.07, 0.25]}, 
						**table_kwargs}

		image_kwargs = {**{'path': None, 'x': -0.35, 'y': 0.5, 'zoom': 0.08}, 
						**image_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		if title is True:
			ax.title.set_text(title_string + ' {0} - {1} \$/kg'.format(self.target_price_range[0], 
																       self.target_price_range[1]))

		yhist, xhist, rectangle = ax.hist(self.distances, bins = bins, density=False, 
										  color=self.color, edgecolor = 'black')
		density = np.sum(np.diff(xhist) * yhist)
	
		mu, std = normal_distribution.fit(self.distances)

		ax.set_xlim(0, 1)
		xmin, xmax = plt.xlim()
		x = np.linspace(xmin, xmax, 500)
		p = normal_distribution.pdf(x, mu, std)
		ax.plot(x, p * density, 'k', linewidth=2)

		if xlabel:
			ax.set_xlabel(xlabel_string)
		ax.set_ylabel(ylabel_string)

		if image_kwargs['path'] is not None:
			insert_image(ax = ax, **image_kwargs)

		self.render_parameter_table(ax, **table_kwargs)
	
		if figure_lean is True:
			figure.execute()
			return figure.fig
								
	def plot_distance_cost_relationship(self, ax = None, ylim = None, figure_lean = True, 
										parameter_table = True, 
	 									legend_loc = 'upper left',
	 									log_scale = False, 
	 									xlabel_string = 'Normalized Distance',
	 									ylabel_string = r'Levelized $H_{2}$ Cost / \$/kg',
	 									table_kwargs = {}, image_kwargs = {}, plot_kwargs = {},
	 									**kwargs):
		'''Plotting relationship of development distance and H2 cost.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		ylim : float, optional
			Limit value for y axis, deactivated when log scale is used. Deafault is None.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		parameter_table : bool, optional
			If parameter_table is True, the parameter table is shown in the plot.
		legend_loc : str, optional
			Controls location of legend in plot. Defaults to 'upper left'.
		log_scale : bool, optional
			If log_scale is True, the y axis will use a log scale.
		x_label_string : str, optional
			String for x axis label.
		y_label_string : str, optional
			String for y axis label.
		table_kwargs : dict, optional
			Dictionary containing optional keyword arguments for 
			:func:`~pyH2A.Analysis.Monte_Carlo_Analysis.Monte_Carlo_Analysis.render_parameter_table`
		image_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.insert_image`
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.
		'''

		kwargs = {**{'left': 0.1, 'right': 0.9, 'bottom': 0.1, 'top': 0.95,
	 			     'fig_width': 7, 'fig_height': 4, 'font_size': 12,
	 			     'name': 'Monte_Carlo_Distance_Cost_Relationship'}, 
	 			  **kwargs, **plot_kwargs}

		table_kwargs = {**{'xpos': 1.4, 'height': 0.23, 'edge_padding': 0.0}, 
						**table_kwargs}

		image_kwargs = {**{'path': None, 'x': 0.5, 'y': 0.8, 'zoom': 0.08}, 
						**image_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		ax.plot(self.results_distances_sorted[:,-1], self.results_distances_sorted[:,-2], '.', 
				markersize = 1, color = self.color, alpha = 0.2)
		ax.plot(self.distances_cost_savgol[:,0], self.distances_cost_savgol[:,1], color = self.color, 
			label = self.display_name)

		if log_scale:
			ax.set_yscale('log')
			fmt = mticker.StrMethodFormatter("{x:g}")
			ax.yaxis.set_major_formatter(fmt)
		else:
			if ylim is not None:
				ax.set_ylim(0.1, ylim)

		ax.set_xlabel(xlabel_string)
		ax.set_ylabel(ylabel_string)
		ax.grid(color = 'grey', linestyle = '--', linewidth = 0.2, zorder = 0)
		ax.legend(loc = legend_loc)

		if parameter_table is True:
			self.render_parameter_table(ax, **table_kwargs)

		if image_kwargs['path'] is not None:
			insert_image(ax = ax, **image_kwargs)

		if figure_lean is True:
			figure.execute()
			return figure.fig




import copy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from pyH2A.Discounted_Cash_Flow import Discounted_Cash_Flow
from pyH2A.Utilities.input_modification import convert_input_to_dictionary, parse_parameter, get_by_path, set_by_path
from pyH2A.Utilities.output_utilities import make_bold, Figure_Lean

class Waterfall_Analysis:
	'''Perform waterfall analysis to study the compounded effect 
	of changing multiple parameters.

	Parameters
	----------
	Waterfall_Analysis > [...] > Name : str
		Display name for parameter, e.g. used for plot labels.
	Waterfall_Analysis > [...] > Type : str
		Type of parameter values. If `Type` is 'value', provided values are
		used as is. If `Type` is 'factor', provided values are multiplied 
		with base value of parameter in input file.
	Waterfall_Analysis > [...] > Value : float
		New value or factor for parameter.
	Waterfall_Analysis > [...] > Show Percent : bool or str, optional
		If there is any entry for `Show Percent` the parameter
		will be displayed as a percentage value in waterfall chart.

	Notes
	-----
	`Waterfall_Analysis` contains parameters which are to be varied in
	waterfall analysis. First column specifies path to parameter in input file 
	(top key > middle key > bottom key format, e.g. Catalyst > Cost per kg ($) > Value).
	Order of varied parameters determines in which order they are applied.
	In the order they are provided, each parameter is changed to the provided value. 
	The relative change of introducing each change is computed,
	and the new H2 cost (compound result of applying all changes) is calculated.
	'''

	def __init__(self, input_file):
		self.inp = convert_input_to_dictionary(input_file)
		self.base_case = Discounted_Cash_Flow(input_file, print_info = False)
		self.results = self.perform_waterfall_analysis()

	def perform_waterfall_analysis(self):
		'''Perform waterfall analysis
		'''

		waterfall = self.inp['Waterfall_Analysis']

		results = {}
		results['Base Case'] = {'Value': self.base_case.h2_cost}

		for i in range(len(waterfall)):
			keys = list(waterfall)[0:i+1]
			dic = {k: waterfall.get(k, None) for k in (keys)}

			self.modify_inp_run_dcf(self.inp, dic, results)

		keys = list(results)[1:] # all keys except first one (base case)

		for key in keys:
			preceding_name = results[key]['Preceding Name']
			preceding_value = results[preceding_name]['Value']
			relative_change = results[key]['Value'] - preceding_value

			results[key]['Preceding Value'] = preceding_value
			results[key]['Relative Change'] = relative_change

		return results

	def modify_inp_run_dcf(self, inp, dic, output):
		'''Modification of `inp` with values from `dic` and 
		running discounted cash flow analysis.
		'''

		inp_modified = copy.deepcopy(inp)

		variable = dic[list(dic)[-1]]['Name']
		output[variable] = {}
		output[variable]['Previous Changes'] = {}

		for key in dic:
			parameters = parse_parameter(key)
			name = dic[key]['Name']
			value = dic[key]['Value']
			value_type = dic[key]['Type']

			base_value = get_by_path(inp_modified, parameters)

			set_by_path(inp_modified, parameters, value, value_type = value_type)

			if dic[key]['Type'] == 'factor':
				shown_value = '{0}x'.format(value)
			else:
				shown_value = self.show_percent(value, dic[key])

			output[variable]['Previous Changes'][name] = shown_value					

		dcf = Discounted_Cash_Flow(inp_modified, print_info = False)

		output[variable]['Value'] = dcf.h2_cost
		output[variable]['Shown Value'] = shown_value
		output[variable]['Base Value'] = self.show_percent(base_value, dic[key])

		try:
			output[variable]['Preceding Name'] = dic[list(dic)[-2]]['Name']
		except IndexError:
			output[variable]['Preceding Name'] = 'Base Case'

	def show_percent(self, value, dic):
		'''Displaying provided `value` as percentage if
		`Show Percent` is in dictionary.
		'''
	
		if 'Show Percent' in dic:
			return '{0}%'.format(value * 100)
		else:
			return value

	def plot_waterfall_chart(self, ax = None, figure_lean = True,
							 width = 0.7, connection_width = 1.0, 
							 label_offset = 20, plot_sorted = False, 
							 plot_kwargs = {},
							 **kwargs):
		'''Plot waterfall chart.

		Parameters
		----------
		ax : matplotlib.axes, optional
			Axes object in which plot is drawn. Default is None, creating new plot.
		figure_lean : bool, optional
			If figure_lean is True, matplotlib.fig object is returned.
		width : float, optional
			Width of bars in waterfall chart.
		connection_width : float, optional
			Width of lines connecting bars.
		label_offset : float, optional
			Offset of label for bars.
		plot_sorted : bool, optional
			If `plot_sorted` is True, bars are plotted in a sorted manner. If it
			False, they are plotted in the order they are provided in the 
			input file. 
		plot_kwargs: dict, optional
			Dictionary containing optional keyword arguments for
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`, has priority over `**kwargs`.
		**kwargs: 
			Additional `kwargs` passed to 
			:func:`~pyH2A.Utilities.output_utilities.Figure_Lean`

		Returns 
		-------
		figure : matplotlib.fig or None
			matplotlib.fig is returned if `figure_lean` is True.
		'''

		results = self.results

		kwargs = {**{'left': 0.12, 'right': 0.95, 'bottom': 0.3, 'top': 0.95,
				     'fig_width': 0.5 * len(results) + 4, 'fig_height': 5,
				     'name': 'Waterfall_Chart'}, 
				  **kwargs, **plot_kwargs}

		if ax is None:
			figure = Figure_Lean(**kwargs)
			ax = figure.ax

		cmap = plt.get_cmap('plasma')

		df = pd.DataFrame.from_dict(results)
		max_value = df.loc['Value'].max(skipna = True)

		label_offset = max_value / label_offset

		if plot_sorted is True:
			df.sort_values(by = ['Relative Change'], axis = 1, 
						   na_position = 'first', inplace = True)
		
		current_value = df['Base Case']['Value']

		labels = []

		for counter, i in enumerate(df):
			if i == 'Base Case':
				bar = patches.Rectangle((counter - width/2, 0), width, df[i]['Value'], 
									     edgecolor = 'none', facecolor = cmap(0.8), 
									     zorder = 5)
				ax.add_patch(bar)
				ax.plot([counter, counter], [0, df[i]['Value']], 
					    color = 'black', linewidth = 0)

				connection = ax.annotate('', xy =(counter + width/2, current_value), 
										 xytext=(counter + 1 - width/2, current_value), 
										 arrowprops={'arrowstyle': '-', 'ls': 'dashed', 
										             'color': 'grey', 
										             'linewidth': connection_width})

				ax.annotate('${:.2f}'.format(df[i]['Value']), 
					         xy = (counter, df[i]['Value'] + label_offset), 
					         va = 'center', ha = 'center')

				labels.append(make_bold(i).replace(' ', '\n'))

			else:
				if df[i]['Relative Change'] >= 0:
					color = 'darkred'
					y = current_value
					y_label = current_value + df[i]['Relative Change'] + label_offset
					height = df[i]['Relative Change']
					
				else:
					color = 'darkgreen'
					y = current_value + df[i]['Relative Change']
					y_label = current_value + df[i]['Relative Change'] - label_offset
					height = abs(df[i]['Relative Change'])

				x = counter - width/2
		
				bar = patches.Rectangle((x, y), width, height, edgecolor = 'none', 
					                    facecolor = color, zorder = 5)
				ax.add_patch(bar)
				ax.plot([counter, counter], 
					    [current_value, current_value + df[i]['Relative Change']], 
					    color = color, linewidth = 0)
	
				ax.annotate('${:.2f}'.format(df[i]['Relative Change']), 
					         xy = (counter, y_label), va = 'center', 
					         ha = 'center', color = color)

				current_value += df[i]['Relative Change']
				connection = ax.annotate('', xy =(counter + width/2, current_value), 
					                     xytext=(counter + 1 - width/2, current_value), 
										arrowprops={'arrowstyle': '-', 'ls': 'dashed', 
										            'color': 'grey', 
										            'linewidth': connection_width})

				labels.append(make_bold(str(df[i]['Shown Value'])) + 
					          '\n' + i.replace(' ', '\n') + '\n' + 
					          make_bold('Base:') + '\n{0}'.format(df[i]['Base Value']))

		bar = patches.Rectangle((counter+1 - width/2, 0), width, 
			                     current_value, edgecolor = 'none', 
			                     facecolor = cmap(0.05), zorder = 5)

		ax.add_patch(bar)
		ax.plot([counter+1, counter+1], [0, current_value], 
			     color = 'black', linewidth = 0)

		ax.annotate('${:.2f}'.format(current_value),
		            xy = (counter+1, current_value + label_offset), 
		            va = 'center', ha = 'center')
		labels.append(make_bold('Adjusted'))

		ax.set_ylabel(r'Levelized cost / USD per kg $H_{2}$')

		ax.set_xticks(np.arange(0, len(results)+1))
		ax.set_xticklabels(labels)

		ylim = np.asarray(ax.get_ylim())
		ylim[1] += 1.2*label_offset
		ylim[0] = 0.0
		ax.set_ylim(ylim)

		ax.grid(color = 'grey', linestyle = '--', linewidth = 0.2, zorder = 0)

		if figure_lean is True:
			figure.execute()
			return figure.fig

		


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('../src/'))

from pyH2A import __version__


# -- Project information -----------------------------------------------------

project = 'pyH2A'
copyright = '2021, Jacob Schneidewind'
author = 'Jacob Schneidewind'

# The full version, including alpha/beta/rc tags
release = __version__
version = __version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
			  'sphinx.ext.autosummary',
   			  'sphinx.ext.todo',
    		  'sphinx.ext.coverage',
    		  'numpydoc',
    		  'autodocsumm'
]

autodoc_default_options = {
    'autosummary': True,
}

numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

source_suffix = '.rst'

master_doc = 'index'

pygments_style = 'sphinx'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

todo_include_todos = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'furo'  #sphinx_rtd_theme

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".

html_static_path    = ['style']        # folders to include in output
html_css_files      = ['custom.css']   # extra style files to apply
html_logo = './_static/pyH2A.svg'





import copy
import numbers
from functools import lru_cache
import numpy as np

from pyH2A.Utilities.input_modification import convert_input_to_dictionary, process_input, process_table, insert, read_textfile, set_by_path, execute_plugin
import pyH2A.Utilities.find_nearest as fn

def numpy_npv(rate, values):
	'''Calculation of net present value.
	'''

	values = np.asarray(values)
	return (values / (1+rate)**np.arange(0, len(values))).sum(axis=0)

@lru_cache(maxsize = None)
def get_idx(diagonal_number, axis0, axis1):
	'''Calculation of index for MACRS calculation.
	Uses ``lru_cache`` for repeated calculations.
	'''

	a = np.arange(0, diagonal_number)
	b = a[::-1]
	c = np.c_[a, b]

	idx = c[(c[:,0] <= axis0 - 1) & (c[:,1] <= axis1 - 1)]
	idx = (np.array(idx[:,0]), np.array(idx[:,1]))

	return idx

def MACRS_depreciation(plant_years, depreciation_length, annual_depreciable_capital):
	'''Calculation of MACRS depreciations.
	
	Parameters
	----------
	plant_years : ndarray
		Array of plant years.
	depreciation_length : int
		Depreciation length.
	annual_depreicable_capital : ndarray
		Depreciable capital by year.

	Returns 
	-------
	annual_charge : ndarray
		Charge by year.

	'''

	end_idx = len(plant_years)	

	original_macrs = read_textfile('pyH2A.Lookup_Tables~MACRS.csv', delimiter = '	')
	macrs = np.copy(original_macrs)

	macrs[1:,1:] = macrs[1:,1:]/100.
	idx_macrs = fn.find_nearest(macrs[0][1:], depreciation_length)[0] 
	macrs_values = macrs[1:,1:][:,idx_macrs]
	macrs_values = macrs_values[macrs_values != 0]

	depreciation = np.outer(annual_depreciable_capital, macrs_values)

	charge = []
	diagonals = sum(depreciation.shape) + 1

	for i in range(1, diagonals):
		idx = get_idx(i, depreciation.shape[0], depreciation.shape[1])

		diagonal = depreciation[idx]
		charge.append(np.sum(diagonal))

	charge = np.asarray(charge)

	annual_charge = charge[:end_idx]
	annual_charge[-1] += np.sum(charge[end_idx:])

	return annual_charge

def discounted_cash_flow_function(inp, values, parameters, attribute = 'h2_cost', 
											plugin = None, plugin_attr = None):
	'''Wrapper function for ``Discounted_Cash_Flow``, substituting provided values 
	at specified parameter positions and returning desired attribute of 
	``Discounted_Cash_Flow`` object.

	Parameters
	----------
	inp : dict or str
		Dictionary containing input information. If `inp` is a file path, the provided 
		file is converted to a dictionary using ``convert_input_to_dictionary``.
	values : ndarray
		1D (in case of one parameter) or 2D array (in case of multiple parameters)
		containing the values which are to be used.
	parameters : ndarray
		1D or 2D array containing the parameter specifications (location within inp);
		Format: [top_key, middle_key, bottom_key].
	attribute : str, optional
		Desired attribute of ``Discounted_Cash_Flow`` object, which should be returned.
		If the attribute is `plugs`, the `.plugs` dictionary attribute is accessed, which 
		contains information of all used plugins (see `plugin` and `plugin_attr`).
		Defaults to `h2_cost`.
	plugin : str, optional
		If `attribute` is set to `plugs`, a `plugin` has to be specified, which should be 
		accessed. Furthermore, a corresponding attribute of the plugin needs to be provided,
		see `plugin_attr`.
	plugin_attr : str, optional
		If `attribute` is set to `plugs`, `plugin_attr` controls which attribute of the 
		specified `plugin` is accessed.

	Returns
	-------
	results : ndarray
		For each value (1D array) or set of values (2D array), the values are 
		substituted in inp, Discounted_Cash_Flow (dcf) is executed and the dcf 
		object is generated. Then, the requested attribute is stored in results, 
		which is finally returned.
	'''

	if isinstance(inp, str):
		inp = convert_input_to_dictionary(inp)

	results = []

	for value_set in values:
		input_dict = copy.deepcopy(inp)

		if isinstance(value_set, numbers.Number):
			set_by_path(input_dict, parameters, value_set)

		else:
			for value, parameter in zip(value_set, parameters):
				set_by_path(input_dict, parameter, value)

		dcf = Discounted_Cash_Flow(input_dict, print_info = False)

		result = getattr(dcf, attribute)

		if attribute == 'plugs':
			result = result[plugin]
			result = getattr(result, plugin_attr)

		results.append(result)

	return results

class Discounted_Cash_Flow:
	'''Class to perform discounted cash flow analysis.

	Parameters
	----------
	Workflow > initial_equity_depreciable_capital > Type : function
		Initial equity depreciable capital function.
	Workflow > initial_equity_depreciable_capital > Position : int
		Position of initial equity depreciable capital function.
	Workflow > non_depreciable_capital_costs > Type : function
		Non-depreciable capital costs function.
	Workflow > non_depreciable_capital_costs > Position : int
		Position of non-depreciable capital costs function.
	Workflow > replacement_costs > Type : function
		Replacement costs function.
	Workflow > replacement_costs > Position : int
		Position of replacement costs function.	
	Workflow > fixed_operating_costs > Type : function
		Fixed operating costs function.
	Workflow > fixed_operating_costs > Position : int
		Position of fixed operating costs function.
	Workflow > variable_operating_costs > Type : function
		Variable operating costs function.
	Workflow > variable_operating_costs > Position : int
		Position of variable operating costs function.
	Workflow > [...] > Type : plugin, optional
		Plugin to be executed.
	Workflow > [...] > Position : int, optional
		Position of plugin to be executed.
	Technical Operating Parameters and Specifications > Output per Year at Gate > Value : float
		Output per year at gate in kg.
	Financial Input Values > ref year > Value : int
		Financial reference year.
	Financial Input Values > startup year > Value : int
		Startup year for plant.
	Financial Input Values > basis year > Value : int
		Financial basis year.
	Financial Input Values > current year capital costs > Value : int
		Current year for capital costs.
	Financial Input Values > startup time > Value : int
		Startup time in years. 
	Financial Input Values > plant life > Value : int
		Plant life in years.
	Financial Input Values > analysis period > Value : int
		Analysis period in years.
	Financial Input Values > depreciation length > Value : int
		Depreciation length in years.
	Financial Input Values > depreciation type > Value : str
		Type of depreciation, currently only MACRS is implemented.
	Financial Input Values > equity > Value : float
		Percentage of equity financing.
	Financial Input Values > interest > Value : float
		Interest rate on debt.
	Financial Input Values > debt > Value : str
		Debt period, currently only constant debt is implemented.
	Financial Input Values > startup cost fixed > Value : float
		Percentage of fixed operating costs during start-up.
	Financial Input Values > startup revenues > Value : float
		Percentage of revenues during start-up.
	Financial Input Values > startup cost variable > Value : float
		Percentage of variable operating costs during start-up.
	Financial Input Values > decommissioning > Value : float
		Decomissioning cost in percentage of depreciable capital investment.
	Financial Input Values > salvage > Value : float
		Salvage value in percentage of total capital investment.
	Financial Input Values > inflation > Value : float
		Inflation rate.
	Financial Input Values > irr > Value : float
		After tax real internal rate of return.
	Financial Input Values > state tax > Value : float
		State tax.
	Financial Input Values > federal tax > Value : float
		Federal tax.
	Financial Input Values > working capital > Value : float
		Working capital as percentage of yearly change in operating costs.
	Construction > [...] > Value : float
		Percentage of capital spent in given year of construction. Number of entries 
		determines construction period in year (each entry corresponds to one year).
		Have to be in order (first entry corresponds to first construction year etc.)
		Values of all entries have to sum to 100%.
	Depreciable Capital Costs > Inflated > Value : float
		Inflated depreciable capital costs.
	Non-Depreciable Capital Costs > Inflated > Value : float
		Inflated non-depreciable capital costs.
	Replacement > Total > Value : float
		Total replacement costs.
	Fixed Operating Costs > Total > Value : float
		Total fixed operating costs.
	Variable Operating Costs > Total > Value : float
		Total variable operating costs.

	Returns
	-------
	Discounted_Cash_Flow : object
		Discounted cash flow analysis object.

	Attributes
	----------
	h2_cost : float
		Levelized H2 cost per kg.
	contributions : dict
		Cost contributions to H2 price.
	plugs : dict 
		Dictionary containing plugin class objects used during analysis.

	Notes
	-----
	**Numerical inputs**

	Numbers use decimal points. Commas can be used as thousands seperator, they are removed from numbers
	during processing. the "%" can be used after a number, indicating that it will be divided by 100
	before being used. 

	**Special symbols**

	Tables in the input file are formatted using GitHub flavoured Markdown. 
	This means that "#" at the beginning of a line indicates a header.
	"|" is used to seperate columns.
	"---" is used on its own line to seperate table headers from table entries.

	Paths to locations in the input file/in self.inp are specified using ">". Paths are always composed
	of three levels: top key > middle key > bottom key.

	File name paths are specified using "/".

	In cases where multiple numbers are used in one field (e.g. during sensitivity analysis), these numbers
	are seperated using ";".

	**Order in the input file**

	Order matters in the following cases: 

	1. For a group of ``sum_all_tables()`` processed tables (sharing the specified part of their key, e.g. "Direct
	Capital Cost"), they are processed in their provided order.
	
	2. Within a table, the first column will be used to generate the "middle key" of self.inp. The order of the
	other columns is not important.

	**Input**

	Processed input cells can contain either a number or path(s) (if multiple paths are used, they have to
	be seperated by ";") to other cells. The use of process_input() (and hence, process_table(), sum_table() and 
	sum_all_tables()) also allows for the value of an input cell to be multiplied by another cell by including
	path(s) in an additiona column (column name typically "Path").

	**Workflow**

	Workflow specifies which functions and plugins are used and in which order they are executed. The listed
	five functions have to be executed in the specified order for pyH2A to work. Plugins can be inserted at 
	appropiate positions (Type: "plugin"). Plugins have to be located in the ./Plugins/ directory. Execcution
	order is determined by the "Position" input. If the specified position is equal to an already exisiting one,
	the function/plugin will be executed after the already specified one. If multiple plugins/function are
	specified with the same position, they will be executed in the order in which they are listed in the 
	input file.

	**Plugins**

	Plugins needs to be composed of a class with the same name as the plugin file name. This class uses two inputs, 
	a discounted cash flow object (usually indicated by "self") and "print_info", which controls the printing of run time
	statements. Within the __init__ function of the class the actions of the plugins are specified, typically call(s)
	of the "insert()" function to modify the discounted cash flow object's "inp" dictionary (self.inp).
	'''

	def __init__(self, input_file, print_info = True, check_processing = True):

		if isinstance(input_file, str):
			self.inp = convert_input_to_dictionary(input_file)
		else:
			self.inp = input_file

		self.print_info = print_info

		process_table(self.inp, 'Financial Input Values', 'Value')
		self.fin = self.inp['Financial Input Values']

		self.npv_dict = {}	
		self.plugs = {}

		self.time()
		self.inflation()

		self.workflow(self.inp, self.npv_dict, self.plugs)  # execution of all functions and plugins specified in "Workflow"

		self.npv_dict['salvage'], self.npv_dict['decomissioning'] = self.salvage_decommissioning()		
		self.npv_dict['working_capital_reserve'] = self.working_capital_reserve_calc()
		self.npv_dict['interest'], self.npv_dict['principal_payment'] = self.debt_financing()
		self.npv_dict['depreciation_charge'] = self.depreciation_charge()
		self.npv_dict['h2_sales'] = self.h2_sales()
		self.h2_cost()
		self.npv_dict['revenue'] = self.h2_revenue()
		self.npv_dict['pre_depreciation_income'], self.npv_dict['taxable_income'], self.npv_dict['taxes'], self.npv_dict['after_tax_income'] = self.income()
		self.cash_flow()
		self.cost_contribution()

		if check_processing is True:
			self.check_processing()

	def workflow(self, inp, npv_dict, plugs_dict):
		'''Executing plugins and functions for discounted cash flow.
		'''

		sorted_keys = sorted(inp['Workflow'], key = lambda x: inp['Workflow'][x]['Position'])

		for key in sorted_keys:
			if inp['Workflow'][key]['Type'] == 'function':
				self.execute_function(key, npv_dict)
			else:
				execute_plugin(key, plugs_dict, print_info = self.print_info, dcf = self)
	
	def execute_function(self, function_name, npv_dict):
		'''Execute class function named `function_name` and store output in `npv_dict`'''

		called_function = getattr(self, function_name)
		output = called_function()
		npv_dict[function_name] = output

	def time(self):
		'''Creating time scale information for discounted cash flow analysis.
		'''

		insert(self, 'Financial Input Values', 'construction time', 'Value', 
			   len(self.inp['Construction']), __name__, print_info = self.print_info)

		construction_start = self.fin['startup year']['Value'] - self.fin['construction time']['Value']
		end_of_life = self.fin['startup year']['Value'] + self.fin['plant life']['Value']

		self.years = np.arange(construction_start, end_of_life)
		self.analysis_years = np.arange(0, self.fin['construction time']['Value'] + 
										   self.fin['plant life']['Value'])
		self.plant_years = np.arange(-self.fin['construction time']['Value'], 
									  self.fin['plant life']['Value'])
		self.operation_years = np.arange(0, self.fin['plant life']['Value'])

	def inflation(self):
		'''Calculate inflation correction and inflators for specific commodities.
		'''

		inflation_rate = 1 + self.fin['inflation']['Value']
		self.inflation_factor = inflation_rate ** self.plant_years
		self.inflation_correction = inflation_rate ** (self.fin['startup year']['Value'] - 
			      									   self.fin['ref year']['Value'])

		plant_cost = read_textfile('pyH2A.Lookup_Tables~Plant_Cost_Index.csv', 
			  						delimiter = '	')
		gdp_deflator_price = read_textfile('pyH2A.Lookup_Tables~GDP_Implicit_Deflator_Price_Index.csv', 
											delimiter = '	')
		labor_price = read_textfile('pyH2A.Lookup_Tables~Labor_Index.csv', 
									 delimiter = '	')
		chemical_price = read_textfile('pyH2A.Lookup_Tables~SRI_Chemical_Price_Index.csv', 
										delimiter = '	')

		plant_idx = fn.find_nearest(plant_cost, [self.fin['current year capital costs']['Value'], 
												 self.fin['basis year']['Value']])
		gdp_idx = fn.find_nearest(gdp_deflator_price, [self.fin['ref year']['Value'], 
													   self.fin['current year capital costs']['Value']])
		labor_idx = fn.find_nearest(labor_price, [self.fin['ref year']['Value'], 
												  self.fin['basis year']['Value']])
		chemical_idx = fn.find_nearest(chemical_price, [self.fin['ref year']['Value'], 
														self.fin['basis year']['Value']])

		self.cepci_inflator = plant_cost[:,1][plant_idx[0]]/plant_cost[:,1][plant_idx[1]]
		self.ci_inflator = gdp_deflator_price[:,1][gdp_idx[0]]/gdp_deflator_price[:,1][gdp_idx[1]]
		self.combined_inflator = self.cepci_inflator * self.ci_inflator
		self.labor_inflator = labor_price[:,1][labor_idx[0]]/labor_price[:,1][labor_idx[1]]
		self.chemical_inflator = chemical_price[:,1][chemical_idx[0]]/chemical_price[:,1][chemical_idx[1]]

	def production_scaling(self):
		'''Get plant outpuer per year at gate.
		'''

		self.output_per_year_at_gate = process_input(self.inp, 
										'Technical Operating Parameters and Specifications', 
										'Output per Year at Gate', 
										'Value')

		return 0.

	def initial_equity_depreciable_capital(self):
		'''Calculate initial equity depreciable capital.
		'''

		self.depreciable_capital = process_input(self.inp, 'Depreciable Capital Costs', 'Inflated', 'Value')
		self.depreciable_capital_inflation = self.depreciable_capital * self.inflation_correction

		process_table(self.inp, 'Construction', 'Value')

		construction_years = []
		for counter, key in enumerate(self.inp['Construction']):
			cost = self.inp['Construction'][key]['Value'] * self.fin['equity']['Value'] * self.depreciable_capital_inflation * self.inflation_factor[counter]
			construction_years.append(cost)

		self.initial_depreciable_capital = np.sum(construction_years)

		self.annual_initial_depreciable_capital = np.zeros(len(self.inflation_factor))
		self.annual_initial_depreciable_capital[:self.fin['construction time']['Value']] = construction_years

		self.after_tax_nominal_irr = (1 + self.fin['irr']['Value']) * (1 + self.fin['inflation']['Value']) - 1

		return numpy_npv(self.after_tax_nominal_irr, construction_years)

	def non_depreciable_capital_costs(self):
		'''Calculate non-depreciable capital costs.
		'''
		
		self.non_depreciable_capital = process_input(self.inp, 'Non-Depreciable Capital Costs', 'Inflated', 'Value')
		self.non_depreciable_capital_inflated = self.non_depreciable_capital * self.inflation_correction
		non_depreciable_capital_inflation_corrected = self.non_depreciable_capital_inflated * self.inflation_factor[0]

		self.annual_non_depreciable_capital = np.zeros(len(self.inflation_factor))
		self.annual_non_depreciable_capital[0] = non_depreciable_capital_inflation_corrected

		return non_depreciable_capital_inflation_corrected

	def replacement_costs(self):
		'''Calculate replacement costs.
		'''
	
		yearly_costs = self.inp['Replacement']['Total']['Value']

		self.start_idx = fn.find_nearest(self.plant_years, 0)[0]
		yearly_costs[:self.start_idx] = 0
		self.annual_replacement_costs = yearly_costs	

		return numpy_npv(self.after_tax_nominal_irr, yearly_costs)

	def salvage_decommissioning(self):
		'''Calculate salvage and decomissioning costs.
		'''

		self.total_capital_inflated = self.depreciable_capital_inflation + self.non_depreciable_capital_inflated

		decommissioning = self.depreciable_capital_inflation * self.fin['decommissioning']['Value']
		salvage = self.total_capital_inflated * self.fin['salvage']['Value']

		self.decommissioning_costs = np.zeros(len(self.plant_years))
		self.decommissioning_costs[-1] = decommissioning * self.inflation_factor[-1]

		self.salvage_income = np.zeros(len(self.plant_years))
		self.salvage_income[-1] = salvage * self.inflation_factor[-1]

		return numpy_npv(self.after_tax_nominal_irr, self.salvage_income), numpy_npv(self.after_tax_nominal_irr, self.decommissioning_costs)

	def fixed_operating_costs(self):
		'''Calculate fixed operating costs.
		'''

		fixed_operating = process_input(self.inp, 'Fixed Operating Costs', 'Total', 'Value')
		fixed_operating_inflated = fixed_operating * self.inflation_correction

		self.start_up_time_idx = self.start_idx + self.fin['startup time']['Value']

		yearly_costs = fixed_operating_inflated * self.inflation_factor
		yearly_costs[:self.start_up_time_idx] = yearly_costs[:self.start_up_time_idx] * self.fin['startup cost fixed']['Value']
		yearly_costs[:self.start_idx] = 0

		self.fixed_operating_costs = yearly_costs

		return numpy_npv(self.after_tax_nominal_irr, yearly_costs)

	def variable_operating_costs(self):
		'''Calculate variable operating costs.
		'''

		variable_operating_costs = self.inflation_factor * self.inp['Variable Operating Costs']['Total']['Value']
		variable_operating_costs[:self.start_up_time_idx] = variable_operating_costs[:self.start_up_time_idx] * self.fin['startup cost variable']['Value']
		variable_operating_costs[:self.start_idx] = 0

		self.variable_operating_costs = variable_operating_costs

		return numpy_npv(self.after_tax_nominal_irr, variable_operating_costs)

	def working_capital_reserve_calc(self):
		'''Calculate working capital reserve.
		'''

		sum_variable_fixed_operating_costs = self.variable_operating_costs + self.fixed_operating_costs

		self.working_capital_reserve = -self.fin['working capital']['Value'] * np.diff(sum_variable_fixed_operating_costs)
		self.working_capital_reserve[-1] = -np.sum(self.working_capital_reserve[:-1])
		self.working_capital_reserve = np.r_[np.zeros(1), self.working_capital_reserve]

		return -numpy_npv(self.after_tax_nominal_irr, self.working_capital_reserve)

	def debt_financing(self):
		'''Calculate constant debt financing.
		'''

		self.debt_financed_capital = self.depreciable_capital_inflation * (1 - self.fin['equity']['Value']) * self.inflation_factor[0]
		interest = self.debt_financed_capital * self.fin['interest']['Value']
		self.interest_per_year = np.ones(len(self.inflation_factor)) * interest

		self.principal_payment = np.zeros(len(self.inflation_factor))
		self.principal_payment[-1] = self.debt_financed_capital

		return numpy_npv(self.after_tax_nominal_irr, self.interest_per_year), numpy_npv(self.after_tax_nominal_irr, self.principal_payment)

	def depreciation_charge(self):
		'''Calculate depreciation charge.
		'''

		total_initial_depreciable_capital = self.debt_financed_capital + self.initial_depreciable_capital
		annual_depreciable_capital = np.copy(self.annual_replacement_costs)
		annual_depreciable_capital[self.start_idx] += total_initial_depreciable_capital

		self.annual_charge = MACRS_depreciation(self.plant_years, self.fin['depreciation length']['Value'], annual_depreciable_capital)		

		return numpy_npv(self.after_tax_nominal_irr, self.annual_charge)

	def h2_sales(self):
		'''Calculate H2 sales.
		'''

		self.annual_sales = np.ones(len(self.inflation_factor)) * self.output_per_year_at_gate
		self.annual_sales[:self.start_up_time_idx] = self.annual_sales[:self.start_up_time_idx] * self.fin['startup revenues']['Value']
		self.annual_sales[:self.start_idx] = 0

		return numpy_npv(self.fin['irr']['Value'], self.annual_sales)

	def h2_cost(self):
		'''Calculate levelized H2 cost.
		'''

		self.total_tax_rate = self.fin['federal tax']['Value'] + self.fin['state tax']['Value'] * (1. - self.fin['federal tax']['Value'])

		lcoe_capital_costs = self.npv_dict['initial_equity_depreciable_capital'] + self.npv_dict['non_depreciable_capital_costs'] + self.npv_dict['replacement_costs'] + self.npv_dict['working_capital_reserve']
		lcoe_depreciation = -self.npv_dict['depreciation_charge'] * self.total_tax_rate
		lcoe_principal_payment = self.npv_dict['principal_payment']
		lcoe_operating_costs = (-self.npv_dict['salvage'] + self.npv_dict['decomissioning'] + self.npv_dict['fixed_operating_costs'] + self.npv_dict['variable_operating_costs'] + self.npv_dict['interest']) * (1. - self.total_tax_rate)
		lcoe_h2_sales = self.npv_dict['h2_sales'] * (1. - self.total_tax_rate)

		self.h2_cost_nominal = (lcoe_capital_costs + lcoe_depreciation + lcoe_principal_payment + lcoe_operating_costs)/lcoe_h2_sales * (1. + self.fin['inflation']['Value']) ** self.fin['construction time']['Value']
		self.h2_cost = self.h2_cost_nominal/self.inflation_correction

	def h2_revenue(self):
		'''Calculate H2 sales revenue.
		'''

		self.annual_revenue = self.annual_sales * self.h2_cost_nominal * self.inflation_factor

		return numpy_npv(self.after_tax_nominal_irr, self.annual_revenue)

	def income(self):
		'''Calculate total income.
		'''

		self.annual_pre_depreciation_income = self.annual_revenue + self.salvage_income - self.decommissioning_costs - self.fixed_operating_costs - self.variable_operating_costs - self.interest_per_year
		self.taxable_income = self.annual_pre_depreciation_income - self.annual_charge
		self.annual_taxes = self.taxable_income * self.total_tax_rate
		self.after_tax_income = self.annual_pre_depreciation_income - self.annual_taxes

		return numpy_npv(self.after_tax_nominal_irr, self.annual_pre_depreciation_income), numpy_npv(self.after_tax_nominal_irr, self.taxable_income), numpy_npv(self.after_tax_nominal_irr, self.annual_taxes), numpy_npv(self.after_tax_nominal_irr, self.after_tax_income)

	def cash_flow(self):
		'''Calculate cash flow.
		'''

		pre_tax_cash_flow = -self.annual_initial_depreciable_capital - self.annual_replacement_costs + self.working_capital_reserve - self.annual_non_depreciable_capital + self.annual_pre_depreciation_income - self.principal_payment
		after_tax_post_depreciation_cash_flow = pre_tax_cash_flow - self.annual_taxes

		npv_after_tax_post_depreciation = numpy_npv(self.after_tax_nominal_irr, after_tax_post_depreciation_cash_flow)

		if abs(npv_after_tax_post_depreciation) > 1e-6:
			print('Warning: NPV of After tax post-depreciation cash flow is not 0, possible error. NPV: {0}'.format(npv_after_tax_post_depreciation))

		cummulative_cash_flow = np.cumsum(after_tax_post_depreciation_cash_flow)

		return numpy_npv(self.after_tax_nominal_irr, cummulative_cash_flow)

	def cost_contribution(self):
		'''Compile contributions to H2 cost.
		'''

		revenue = self.expenses_per_kg_H2(self.npv_dict['revenue'])

		self.contributions = {'Data': {'Initial equity depreciable capital': self.expenses_per_kg_H2(self.npv_dict['initial_equity_depreciable_capital']),
						   			   'Non depreciable capital' : self.expenses_per_kg_H2(self.npv_dict['non_depreciable_capital_costs']),
						 			   'Replacement costs' : self.expenses_per_kg_H2(self.npv_dict['replacement_costs']),
						      		   'Salvage' : -self.expenses_per_kg_H2(self.npv_dict['salvage']),
						 	  		   'Decomissioning' : self.expenses_per_kg_H2(self.npv_dict['decomissioning']),
						  	  		   'Fixed operating costs' : self.expenses_per_kg_H2(self.npv_dict['fixed_operating_costs']),
						 	  		   'Variable operating costs' : self.expenses_per_kg_H2(self.npv_dict['variable_operating_costs']),
						 	  		   'Working capital reserve' : self.expenses_per_kg_H2(self.npv_dict['working_capital_reserve']),
						   	  		   'Interest' : self.expenses_per_kg_H2(self.npv_dict['interest']),
						      		   'Principal payment' : self.expenses_per_kg_H2(self.npv_dict['principal_payment']),
						      		   'Taxes' : self.expenses_per_kg_H2(self.npv_dict['taxes'])}
						      		   }

		self.contributions['Total'] = self.h2_cost
		self.contributions['Table Group'] = 'Total cost of hydrogen'

	def expenses_per_kg_H2(self, value):
		'''Calculate expenses per kg H2.
		'''

		return value/self.npv_dict['h2_sales'] * (1. + self.fin['inflation']['Value']) ** self.fin['construction time']['Value'] / self.inflation_correction

	def check_processing(self):
		'''Check whether all tables in input file were used.

		Notes
		-----
		'Workflow' and 'Display Parameters' tables are exempted.
		Furthermore, all tables that have the term 'Analysis' in their name
		are also exempted.

		'''

		exceptions = ['Workflow', 'Display Parameters']

		for top_key in self.inp:
			if top_key not in exceptions and 'Analysis' not in top_key:
				for middle_key in self.inp[top_key]:
					if 'Processed' not in self.inp[top_key][middle_key]:
						print('Warning: "{0} > {1}" has not been processed'.format(top_key, middle_key))
